# -*- coding: utf-8 -*-
"""
Created on Mon Aug 27 17:08:42 2018

@author: tb267
"""

from . import analysis
from . import file
from . import plotting
from . import modal

import numpy as np
import datetime
import uuid
import copy

#%% version
VERSION = '0.6.4' # keep in sync with setup.py

def update_dataset(dataset):
    dataset_new = DataSet()
    dataset_new.add_to_dataset(dataset.time_data_list)
    dataset_new.add_to_dataset(dataset.freq_data_list)
    dataset_new.add_to_dataset(dataset.tf_data_list)
    dataset_new.add_to_dataset(dataset.cross_spec_data_list)
    dataset_new.add_to_dataset(dataset.sono_data_list)
    dataset_new.add_to_dataset(dataset.meta_data_list)
    if hasattr(dataset,'modal_data_list'):
        dataset_new.add_to_dataset(dataset.modal_data_list)
    else:
        dataset.modal_data_list = ModalDataList()
    for tf_data in dataset_new.tf_data_list:
        if not hasattr(tf_data,'flag_modal_TF'):
            tf_data.flag_modal_TF = False
    return dataset_new
    
#%% Data structure
class DataSet():
    def __init__(self,data=None):#,*,timedata=[],freqdata=[],cspecdata=[],tfdata=[],sonodata=[],metadata=[]):
        ## initialisation function to set up DataSet class
        
        self.time_data_list = TimeDataList()
        self.freq_data_list = FreqDataList()
        self.cross_spec_data_list = CrossSpecDataList()
        self.tf_data_list = TfDataList()
        self.modal_data_list = ModalDataList()
        self.sono_data_list = SonoDataList()
        self.meta_data_list = MetaDataList()
        
        if not data == None:
            self.add_to_dataset(data)
            
        self.pydvma_version = VERSION
            
        
    def add_to_dataset(self,data):
        ## find out what kind of data being added
        ## allow input to be list of single type of data, or unit data class
        if not 'list' in data.__class__.__name__.lower():
            # turn into list even if unit length
            data = [data]
        else:
            # check list contains set of same kind of data
            check = True
            for d in data:
                check = check and (d.__class__.__name__ == data[0].__class__.__name__)
            if check is False:
                raise ValueError('Data list needs to contain homogenous type of data')
        if len(data) != 0:
            data_class = data[0].__class__.__name__    
        else:
            data_class = None
            
        #print('')
        if data_class=='TimeData':
            self.time_data_list += data
            #print('{} added to dataset'.format(data))
        elif data_class=='FreqData':
            self.freq_data_list += data
            #print('{} added to dataset'.format(data))
        elif data_class=='CrossSpecData':
            self.cross_spec_data_list += data
            #print('{} added to dataset'.format(data))
        elif data_class=='TfData':
            self.tf_data_list += data
            #print('{} added to dataset'.format(data))
        elif data_class=='ModalData':
            self.modal_data_list += data
            #print('{} added to dataset'.format(data))
        elif data_class=='SonoData':
            self.sono_data_list += data
            #print('{} added to dataset'.format(data))
        elif data_class=='MetaData':
            self.meta_data_list += data
            #print('{} added to dataset'.format(data))
        else:
            pass#print('No data added')
        
    def replace_data_item(self,data,n_set):
        ## replace a specific data item
        ## useful for replacing logged data
        ## useful for replacing reconstructed modal data
        
        data_class = data.__class__.__name__    
            
        if data_class=='TimeData':
            self.time_data_list[n_set] = data
        elif data_class=='FreqData':
            self.freq_data_list[n_set] = data
        elif data_class=='CrossSpecData':
            self.cross_spec_data_list[n_set] = data
        elif data_class=='TfData':
            self.tf_data_list[n_set] = data
        elif data_class=='ModalData':
            self.modal_data_list[n_set] = data
        elif data_class=='SonoData':
            self.sono_data_list[n_set] = data
        elif data_class=='MetaData':
            self.meta_data_list[n_set] = data
        else:
            pass
        
            
    def remove_last_data_item(self,data_class):
        
        if data_class == 'TimeData':
            if len(self.time_data_list) != 0:
                del self.time_data_list[-1]
        if data_class == 'FreqData':
            if len(self.freq_data_list) != 0:
                del self.freq_data_list[-1]
        if data_class == 'CrossSpecData':
            if len(self.cross_spec_data_list) != 0:
                del self.cross_spec_data_list[-1]
        if data_class == 'TfData':
            if len(self.tf_data_list) != 0:
                del self.tf_data_list[-1]
        if data_class == 'ModalData':
            if len(self.modal_data_list) != 0:
                del self.modal_data_list[-1]
        if data_class == 'SonoData':
            if len(self.sono_data_list) != 0:
                del self.sono_data_list[-1]
        if data_class == 'MetaData':
            if len(self.meta_data_list) != 0:
                del self.meta_data_list[-1]
                
        #print(self)
                
    def remove_data_item_by_index(self,data_class,list_index):
        
        if list_index.__class__.__name__ == 'ndarray':
            list_index = list(list_index)
        elif type(list_index) is int:
            list_index = [list_index]
            
        list_index.sort()

        if data_class == 'TimeData':
            if len(self.time_data_list) > np.max(list_index):
                for i in reversed(list_index):
                    del self.time_data_list[i]
            else:
                print('indices out of range, no data removed')
                
        if data_class == 'FreqData':
            if len(self.freq_data_list) > np.max(list_index):
                for i in reversed(list_index):
                    del self.freq_data_list[i]
            else:
                print('indices out of range, no data removed')
        
        if data_class == 'CrossSpecData':
            if len(self.cross_spec_data_list) > np.max(list_index):
                for i in reversed(list_index):
                    del self.cross_spec_data_list[i]
            else:
                print('indices out of range, no data removed')
                
        if data_class == 'TfData':
            if len(self.tf_data_list) > np.max(list_index):
                for i in reversed(list_index):
                    del self.tf_data_list[i]
            else:
                print('indices out of range, no data removed')
                
        if data_class == 'ModalData':
            if len(self.modal_data_list) > np.max(list_index):
                for i in reversed(list_index):
                    del self.tf_data_list[i]
            else:
                print('indices out of range, no data removed')
                    
        if data_class == 'SonoData':
            if len(self.sono_data_list) > np.max(list_index):
                for i in reversed(list_index):
                    del self.sono_data_list[i]
            else:
                print('indices out of range, no data removed')
                
        if data_class == 'MetaData':
            if len(self.meta_data_list) > np.max(list_index):
                for i in reversed(list_index):
                    del self.meta_data_list[i] 
            else:
                print('indices out of range, no data removed')

        #print(self)
        
    def calculate_fft_set(self,time_range=None,window=None):
        '''
        Calls analysis.calculate_fft on each TimeData item in the TimeDataList and adds FreqDataList object to dataset
        '''
        if len(self.time_data_list)>0:
            freq_data_list = self.time_data_list.calculate_fft_set(time_range=time_range,window=window)
            self.freq_data_list = freq_data_list
            #self.add_to_dataset(freq_data_list)
        else:
            self.freq_data_list = FreqDataList()
            print('No time data found in dataset')
            
    def calculate_tf_set(self, ch_in=0, time_range=None,window=None,N_frames=1,overlap=0.5):
        '''
        Calls analysis.calculate_fft on each TimeData item in the TimeDataList and adds FreqDataList object to dataset
        '''
        if len(self.time_data_list)>0:
            tf_data_list = self.time_data_list.calculate_tf_set(ch_in=ch_in, time_range=time_range, window=window, N_frames=N_frames, overlap=overlap)
            self.tf_data_list = tf_data_list
            #self.add_to_dataset(tf_data_list)
        else:
            self.tf_data_list = TfDataList()
            print('No time data found in dataset')
            
    def calculate_cross_spectrum_matrix_set(self,ch_in=0, time_range=None,window='hann',N_frames=1,overlap=0.5):
        '''
        Calls analysis.calculate_fft on each TimeData item in the TimeDataList and adds FreqDataList object to dataset
        '''
        if len(self.time_data_list)>0:
            cross_spec_data_list = self.time_data_list.calculate_cross_spectrum_matrix_set(ch_in=ch_in, time_range=time_range,window='hann',N_frames=N_frames,overlap=overlap)
            self.cross_spec_data_list = cross_spec_data_list
            #self.add_to_dataset(cross_spec_data_list)
        else:
            self.cross_spec_data_list = CrossSpecDataList()
            print('No time data found in dataset')
            
    def calculate_tf_averaged(self, ch_in=0, time_range=None,window='hann'):
        '''
        Calls analysis.calculate_fft on each TimeData item in the TimeDataList and adds FreqDataList object to dataset
        '''
        if len(self.time_data_list)>0:
            tf_data = self.time_data_list.calculate_tf_averaged(ch_in=ch_in, time_range=time_range ,window=window)
            self.tf_data_list = TfDataList([tf_data])
            #self.add_to_dataset(tf_data)
        else:
            self.tf_data_list = TfDataList()
            print('No time data found in dataset')
            
    def calculate_cross_spectra_averaged(self, time_range=None,window=None):
        '''
        Calls analysis.calculate_fft on each TimeData item in the TimeDataList and adds FreqDataList object to dataset
        '''
        if len(self.time_data_list)>0:
            cross_spec_data = self.time_data_list.calculate_cross_spectra_averaged(time_range=time_range,window=window)
            self.cross_spec_data_list = CrossSpecDataList([cross_spec_data])
            #self.add_to_dataset(cross_spec_data)
        else:
            self.cross_spec_data_list = CrossSpecDataList()
            print('No time data found in dataset')
            
    def calculate_sono_set(self, nperseg=None):
        if len(self.time_data_list)>0:
            sono_data_list = self.time_data_list.calculate_sono_set(nperseg=nperseg)
            self.sono_data_list = sono_data_list
        else:
            self.sono_data_list = SonoDataList()
            print('No time data foudn in dataset')
            
    def clean_impulse(self,ch_impulse=0):
        '''
        Calls analysis.clean_impulse on each TimeData item in the TimeDataList and returns a copy of the new dataset.
        
        Note that calling this function *does not* change the data, and just returns a copy.
        '''
        dataset_copy = copy.deepcopy(self)
        dataset_copy.remove_data_item_by_index('TimeData',np.arange(len(dataset_copy.time_data_list)))
        if len(self.time_data_list)>0:
            for time_data in self.time_data_list:
                td = analysis.clean_impulse(time_data, ch_impulse=ch_impulse)
                dataset_copy.add_to_dataset(td)
            print('returning copy of data with impulses cleaned')
            return dataset_copy
        else:
            print('No time data found in dataset')
            return None
            
    def save_data(self, filename=None):
        savename = file.save_data(self, filename=filename, overwrite_without_prompt=False)
        return savename
    
    def export_to_matlab(self, filename=None, overwrite_without_prompt=False):
        savename = file.export_to_matlab(self, filename=filename, overwrite_without_prompt=overwrite_without_prompt)
        return savename
    
    def export_to_matlab_jwlogger(self, filename=None, overwrite_without_prompt=False):
        savename = file.export_to_matlab_jwlogger(self, filename=filename, overwrite_without_prompt=overwrite_without_prompt)
        return savename
    
    def plot_time_data(self,sets='all',channels='all'):
        global pt
        pt = plotting.PlotData()
        pt.update(self.time_data_list,sets=sets,channels=channels)
        return pt
        
    def plot_freq_data(self,sets='all',channels='all'):
        global pf
        pf = plotting.PlotData()
        pf.update(self.freq_data_list,sets=sets,channels=channels)
        return pf
        
    def plot_tf_data(self,sets='all',channels='all'):
        global ptf
        ptf = plotting.PlotData()
        ptf.update(self.tf_data_list,sets=sets,channels=channels)
        return ptf
    
    def plot_sono_data(self,n_set=0, n_chan=0, db_range=60):
        global ptf
        ptf = plotting.PlotData()
        ptf.update_sonogram(self.sono_data_list,n_set=n_set,n_chan=n_chan, db_range=db_range)
        return ptf
    
    def __repr__(self):
        template = "{:>24}: {}"
        dataset_dict = self.__dict__
        text = '\n<DataSet> class:\n\n'
        for attr in dataset_dict:
            N = len(dataset_dict[attr])
            if N <= 3:
                text += template.format(attr,dataset_dict[attr])
                text += '\n'
            elif attr == 'pydvma_version':
                pass#text += template.format('pydvma_version',str(self.pydvma_version))
            else:
                text += template.format(attr,'[' + str(dataset_dict[attr][0]) + ',... (x' + str(N) + ')]')
                text += '\n'
        
        return text
    
class TimeDataList(list):
    ### This will allow functions to be discovered that can take lists of TimeData is arguments
    def calculate_fft_set(self,time_range=None,window=None):
        '''
        Calls analysis.calculate_fft on each item in the list and returns FreqDataList object
        '''
        freq_data_list = FreqDataList()
        
        for td in self:
            freq_data = analysis.calculate_fft(td, time_range=time_range, window=window)
            freq_data_list += [freq_data]
            
        return freq_data_list
    
    
    def calculate_tf_set(self, ch_in=0, time_range=None,window=None,N_frames=1,overlap=0.5):
        '''
        Calls analysis.calculate_tf on each item in the list and returns TfDataList object
        '''
        tf_data_list = TfDataList()
        
        for td in self:
            tf_data = analysis.calculate_tf(td, ch_in=ch_in, time_range=time_range,window=window,N_frames=N_frames,overlap=overlap)
            tf_data_list += [tf_data]
            
        return tf_data_list
    
    def calculate_cross_spectrum_matrix_set(self, ch_in=0, time_range=None,window=None,N_frames=1,overlap=0.5):
        '''
        Calls analysis.calculate_tf on each item in the list and returns TfDataList object
        '''
        cross_spec_data_list = CrossSpecDataList()
        
        for td in self:
            cross_spec_data = analysis.calculate_cross_spectrum_matrix(td, time_range=time_range,window=window,N_frames=N_frames,overlap=overlap)
            cross_spec_data_list += [cross_spec_data]
            
        return cross_spec_data_list
    
    
    def calculate_tf_averaged(self, ch_in=0, time_range=None,window='hann'):
        '''
        Calls analysis.calculate_tf_averaged on whole list and returns TfData object
        '''
        tf_data = analysis.calculate_tf_averaged(self,ch_in=ch_in, time_range=time_range,window=window)
            
        return tf_data
    
    
    def calculate_cross_spectra_averaged(self, time_range=None,window=None):
        '''
        Calls analysis.calculate_tf_averaged on whole list and returns TfData object
        '''
        cross_spec_data = analysis.calculate_cross_spectra_averaged(self, time_range=time_range,window=window)
            
        return cross_spec_data
    
    def calculate_sono_set(self, nperseg=None):
        '''
        Calls analysis.calculate_sonogram on each item in the list and returns SonoDataList object
        '''
        sono_data_list = SonoDataList()
        
        for td in self:
            sono_data = analysis.calculate_sonogram(td,nperseg=nperseg)
            sono_data_list += [sono_data]
            
        return sono_data_list
    
    def get_calibration_factors(self):
        n_set = len(self)
        factors = []
        for ns in range(n_set):
            factors.append(self[ns].channel_cal_factors)
        
        return factors
            
    def set_calibration_factors_all(self,factors):
        n_set = len(self)
        for ns in range(n_set):
            self[ns].channel_cal_factors=factors[ns]
            
    def set_calibration_factor(self,factor, n_set=0, n_chan=0):
        if len(self) == 0:
            print('<TimeDataList> is empty. First log data, load data, or create test data.')
        elif n_set >= len(self):
            print('<TimeDataList> has {} set(s) of <TimeData>. Set requested (index={}) exceeds number of sets. Note indexing starts at 0.'.format(len(self),n_set))
        elif n_chan >= len(self[n_set].tf_data[0,:]):
            print('<TimeDataList>[{}] has {} channel(s). Channel requested (index={}) exceeds number of channels. Note indexing starts at 0.'.format(n_set,len(self[n_set].time_data[0,:]),n_chan))
        else:
            self[n_set].channel_cal_factors[n_chan]=factor
    
    def export_to_csv(self, filename=None, overwrite_without_prompt=False):
        savename = file.export_to_csv(self,filename=filename,overwrite_without_prompt=overwrite_without_prompt)
        return savename

class FreqDataList(list):
    ### This will allow functions to be discovered that can take lists of FreqData is arguments
    def get_calibration_factors(self):
        n_set = len(self)
        factors = []
        for ns in range(n_set):
            factors.append(self[ns].channel_cal_factors)
        
        return factors
    
    def set_calibration_factors_all(self,factors):
        n_set = len(self)
        for ns in range(n_set):
            self[ns].channel_cal_factors=factors[ns]
            
    def set_calibration_factor(self,factor, n_set=0, n_chan=0):
        if len(self) == 0:
            print('<FreqDataList> is empty. First calculate FFT.')
        elif n_set >= len(self):
            print('<FreqDataList> has {} set(s) of <FreqData>. Set requested (index={}) exceeds number of sets. Note indexing starts at 0.'.format(len(self),n_set))
        elif n_chan >= len(self[n_set].tf_data[0,:]):
            print('<FreqDataList>[{}] has {} channel(s). Channel requested (index={}) exceeds number of channels. Note indexing starts at 0.'.format(n_set,len(self[n_set].freq_data[0,:]),n_chan))
        else:
            self[n_set].channel_cal_factors[n_chan]=factor
            
    def export_to_csv(self, filename=None, overwrite_without_prompt=False):
        savename = file.export_to_csv(self,filename=filename,overwrite_without_prompt=overwrite_without_prompt)
        return savename

class CrossSpecDataList(list):
    ### This will allow functions to be discovered that can take lists of CrossSpecData is arguments
    pass

class TfDataList(list):
    ### This will allow functions to be discovered that can take lists of TfData is arguments
    def get_calibration_factors(self):
        n_set = len(self)
        factors = []
        for ns in range(n_set):
            factors.append(self[ns].channel_cal_factors)
        
        return factors
    
    def set_calibration_factors_all(self,factors):
        n_set = len(self)
        for ns in range(n_set):
            self[ns].channel_cal_factors=factors[ns]
            
    def set_calibration_factor(self,factor, n_set=0, n_chan=0):
        if len(self) == 0:
            print('<TfDataList> is empty. First calculate transfer function.')
        elif n_set >= len(self):
            print('<TfDataList> has {} set(s) of <TfData>. Set requested (index={}) exceeds number of sets. Note indexing starts at 0.'.format(len(self),n_set))
        elif n_chan >= len(self[n_set].tf_data[0,:]):
            print('<TfDataList>[{}] has {} channel(s). Channel requested (index={}) exceeds number of channels. Note indexing starts at 0.'.format(n_set,len(self[n_set].tf_data[0,:]),n_chan))
        else:
            self[n_set].channel_cal_factors[n_chan]=factor
    
    def add_modal_reconstruction(self,tf_data,mode='replace'):
        # identify number of TFs in list that are reconstructions
        N_reconstruction = 0
        for tf in self:
            if tf.flag_modal_TF == True:
                N_reconstruction += 1
                
        # append / replace reconstruction TFs
        if N_reconstruction == 0:
            self += [tf_data]
        elif mode == 'replace':
            self[-1] = tf_data
        elif mode == 'append':
            self += [tf_data]
            
        
    def export_to_csv(self, filename=None, overwrite_without_prompt=False):
        savename = file.export_to_csv(self,filename=filename,overwrite_without_prompt=overwrite_without_prompt)
        return savename
      
class ModalDataList(list):
    ### This will allow functions to be discovered that can take lists of ModalData is arguments
    pass

class SonoDataList(list):
    ### This will allow functions to be discovered that can take lists of SonoData is arguments
    pass

class MetaDataList(list):
    ### This will allow functions to be discovered that can take lists of MetaData is arguments
    pass

        
class TimeData():
    def __init__(self,time_axis,time_data,settings,timestamp=None,timestring=None,units=None,channel_cal_factors=None,id_link=None,test_name=None):
        
        time_data = reshape_arrays(time_data)
        if channel_cal_factors is None:
            channel_cal_factors = np.ones(len(time_data[0,:]))
        
        if timestamp == None:
            t = datetime.datetime.now()
            timestamp = t
            timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
        
        self.time_axis = time_axis
        self.time_data = time_data  
        self.settings = settings
        self.timestamp = timestamp
        self.timestring = timestring
        self.units = units
        self.channel_cal_factors = channel_cal_factors
        self.id_link = id_link # this is used if data is derived from an existing <TimeData> measurement
        self.test_name = test_name
        self.unique_id = uuid.uuid4()
        
        
        
    def __repr__(self):
        return "<TimeData>"

        
class FreqData():
    def __init__(self,freq_axis,freq_data,settings,units=None,channel_cal_factors=None,id_link=None,test_name=None):
        
        freq_data = reshape_arrays(freq_data)
        if channel_cal_factors is None:
            channel_cal_factors = np.ones(len(freq_data[0,:]))
        
        self.freq_axis = freq_axis
        self.freq_data = freq_data
        self.settings = settings
        self.test_name = test_name
        self.units = units
        self.channel_cal_factors = channel_cal_factors
        self.id_link = id_link # used to link data to specific <TimeData> object
        t = datetime.datetime.now()
        self.timestamp = t
        self.timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
        
    def __repr__(self):
        return "<FreqData>"
    
    
class CrossSpecData():
    def __init__(self,freq_axis,Pxy,Cxy,settings,units=None,channel_cal_factors=None,id_link=None,test_name=None):
        
        self.freq_axis = freq_axis
        self.Pxy = Pxy
        self.Cxy = Cxy
        self.settings = settings
        self.test_name = test_name
        self.units = units
        self.channel_cal_factors = channel_cal_factors
        self.id_link = id_link # used to link data to specific <TimeData> object
        t = datetime.datetime.now()
        self.timestamp = t
        self.timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
        
    def __repr__(self):
        return "<CrossSpecData>"
    
        
class TfData():
    def __init__(self,freq_axis,tf_data,tf_coherence,settings,units=None,channel_cal_factors=None,id_link=None,test_name=None):
        
        tf_data = reshape_arrays(tf_data)
        if channel_cal_factors is None:
            channel_cal_factors = np.ones(len(tf_data[0,:]))
        
        self.freq_axis = freq_axis
        self.tf_data = tf_data
        self.tf_coherence = tf_coherence
        self.settings = settings
        self.test_name = test_name
        self.units = units
        self.channel_cal_factors = channel_cal_factors
        self.id_link = id_link # used to link data to specific <TimeData> object
        t = datetime.datetime.now()
        self.timestamp = t
        self.timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
        self.flag_modal_TF = False
        
    def __repr__(self):
        return "<TfData>"
    
    
class ModalData():
    def __init__(self,xn=None,settings=None,units=None,id_link=None,test_name=None):
        
        self.M = []
        self.test_name = test_name
        self.settings = settings
        self.channels = 0
        if settings is not None:
            self.settings.channels = 0
        self.units = units
        self.id_link = id_link # used to link data to specific <TimeData> object
        t = datetime.datetime.now()
        self.timestamp = t
        self.timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)

        if xn is not None:
            self.add_mode(xn)


    def add_mode(self,xn):
        # Make modal matrix. Each row is modal vector stacked as per 'x' in modal.py
        if len(self.M) == 0:
            self.M = np.atleast_2d(xn)
        elif len(xn) == len(self.M[0,:]):
            self.M = np.vstack((self.M,xn))
        else:
            print('Incompatible mode: different number of channels to existing set.')
            return
        
        # sort by frequency (first column)
        sort_i = np.argsort(self.M[:,0])
        self.M = self.M[sort_i,:]
        self.channels = len(sort_i)
        self.settings.channels = self.channels
        
        # separate properties for easier summary, and don't need summary of local residuals rk and rm
        fn,zn,an,pn,rk,rm = modal.unpack_matrix(self.M)
        self.fn = fn
        self.zn = zn
        self.an = an
        self.pn = pn
        
    def delete_mode(self,mode_number):
        self.M = np.delete(self.M,mode_number,0)
        self.channels = len(self.M[:,0])
        self.settings.channels = self.channels
        
        # separate properties for easier summary, and don't need summary of local residuals rk and rm
        self.fn = self.M[:,0]
        self.zn = self.M[:,1]
        self.an = self.M[:,2]
        self.pn = self.M[:,3]
        
            
    def __repr__(self):
        return "<ModalData>"
        
#    def __repr__(self):
#        with np.printoptions(precision=3, suppress=True):
#            template = "{}: {}"
#            modal_dict = self.__dict__
#            text = '\n<ModalData> class:\n\n'
#            for attr in modal_dict:
#                print(attr)
#                if (attr != 'xn') & (attr != 'rk') & (attr != 'rm') & (attr != 'units') & (attr != 'test_name')& (attr != 'id_link')& (attr != 'timestamp')& (attr != 'timestring'):
#                    text += template.format(attr,modal_dict[attr])
#                    text += '\n'
#            
#            return text
    
        
class SonoData():
    def __init__(self,time_axis,freq_axis,sono_data,settings,units=None,channel_cal_factors=None,id_link=None,test_name=None):
        self.time_axis = time_axis
        self.freq_axis = freq_axis
        self.sono_data = sono_data
        self.settings = settings
        self.test_name = test_name
        self.units = units
        self.channel_cal_factors = channel_cal_factors
        self.id_link = id_link # used to link data to specific <TimeData> object
        t = datetime.datetime.now()
        self.timestamp = t
        self.timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
        
    def __repr__(self):
        return "<SonoData>"
        
class MetaData():
    def __init__(self, units=None, channel_cal_factors=None, tf_cal_factors = None,test_name=None):
        ### not sure this is a helpful datafield: might delete. Metadata then contained within each data unit.
        self.units = units
        self.channel_cal_factors = None
        self.tf_cal_factors = None
        t = datetime.datetime.now()
        self.timestamp = t
        self.timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
        
    def __repr__(self):
        return "<MetaData>"
    
    
def reshape_arrays(a):
    b = np.shape(a)
    if len(b) == 1:
        a = a[:,None]
        
    return a


# -*- coding: utf-8 -*-
"""
Created on Sun Sep  2 20:16:26 2018

@author: tb267
"""

from . import datastructure

import numpy as np
from scipy import signal
import copy

MESSAGE = ''


def calculate_fft(time_data,time_range=None,window=None):
    '''
    Args:
        time_data (<TimeData> object): time series data
        time_range: 2x1 numpy array to specify data segment to use
        window (bool): apply blackman filter to data before fft or not
    '''
    
    if time_data.__class__.__name__ != 'TimeData':
        raise Exception('Input data needs to be single <TimeData> object')
    

    if time_range == None:
        ### use all data
        time_range_copy = time_data.time_axis[[0,-1]]
        
    elif time_range.__class__.__name__ == 'PlotData':
        time_range_copy=time_range.ax.get_xbound()
        
    settings = copy.copy(time_data.settings)
    settings.window = window
    settings.time_range = time_range_copy

    s1 = time_data.time_axis >= time_range_copy[0]
    s2 = time_data.time_axis <= time_range_copy[1]
    selection = s1 & s2
    data_selected = time_data.time_data[selection,:]
    N = len(data_selected[:,0])
    
    if window is None:
        w = signal.windows.get_window('boxcar',N)
    else:
        w = signal.windows.get_window(window,N)
        
    # Broadcast the window across potentially multiple channels
    data_selected = w[:, None] * data_selected
        
    fdata = np.fft.rfft(data_selected,axis=0)
    faxis = np.fft.rfftfreq(N,1/time_data.settings.fs)
    
    freq_data = datastructure.FreqData(faxis,fdata,settings,id_link=time_data.unique_id,test_name=time_data.test_name)
    
    return freq_data

def multiply_by_power_of_iw(data,power,channel_list):
    
    if data.__class__.__name__ == 'TfData':
        iw = 1j*2*np.pi * data.freq_axis[:,None]
        if power<0:
            iw[0]=np.inf
        data.tf_data[:,channel_list] = (iw**power) * data.tf_data[:,channel_list]
        # keep track of multiplication powers
        if hasattr(data,'iw_power_counter'):
            data.iw_power_counter[channel_list] += power
        else:
            data.iw_power_counter = np.zeros(len(data.tf_data[0,:]))
            data.iw_power_counter[channel_list] = power
            
    elif data.__class__.__name__ == 'FreqData':
        iw = 1j*2*np.pi * data.freq_axis[:,None]
        if power<0:
            iw[0]=np.inf
        data.freq_data[:,channel_list] = (iw**power) * data.freq_data[:,channel_list]
        if hasattr(data,'iw_power_counter'):
            data.iw_power_counter[channel_list] += power
        else:
            data.iw_power_counter = np.zeros(len(data.freq_data[0,:]))
            data.iw_power_counter[channel_list] = power
    else:
        print('Expecting input argument of type <FreqData> or <TfData>')

            
    return data




def best_match(tf_data_list,freq_range=None,set_ref=0,ch_ref=0):
    '''
    Args:
        tf_data (<TfData> object): transfer function data
        freq_range: 2x1 numpy array to specify data segment to use
    '''
    
    if tf_data_list.__class__.__name__ != 'TfDataList':
        raise ValueError('Input data needs to be single <TfData> object')
    

    if freq_range == None:
        ### use all data
        freq_range_copy = tf_data_list[set_ref].freq_axis[[0,-1]]
        
    elif freq_range.__class__.__name__ == 'PlotData':
        freq_range_copy=freq_range.tfax.get_xbound()
        
    else:
        freq_range_copy = freq_range
        
    
    settings = copy.copy(tf_data_list[0].settings)
    settings.freq_range = freq_range_copy

    
    n_set = len(tf_data_list)
    
    s1 = tf_data_list[set_ref].freq_axis >= freq_range_copy[0]
    s2 = tf_data_list[set_ref].freq_axis <= freq_range_copy[1]
    selection_ref = s1 & s2
    # choose 0-1 scale to make dimensions of data compatible
    f_ref = tf_data_list[set_ref].freq_axis[selection_ref]
        
    factors = []
    for ns in range(n_set):
        f=[]
        n_chan = len(tf_data_list[ns].tf_data[0,:])
        s1 = tf_data_list[ns].freq_axis >= freq_range_copy[0]
        s2 = tf_data_list[ns].freq_axis <= freq_range_copy[1]
        selection = s1 & s2
        f_sel = tf_data_list[ns].freq_axis[selection]
        N_ref = len(f_ref)
        N_sel = len(f_sel)
        f_newref = np.linspace(freq_range[0],freq_range[1],np.max([N_ref,N_sel]))
        
        for nc in range(n_chan):
            # could make more efficient by doing all channels at once
            x = tf_data_list[ns].freq_axis
            y = tf_data_list[ns].tf_data[:,nc]
            a = np.interp(f_newref,x,y)
            a2d = a.reshape(np.size(a),1)
            
            x = tf_data_list[set_ref].freq_axis
            y = tf_data_list[set_ref].tf_data[:,ch_ref]
            b = np.interp(f_newref,x,y)
            b2d = b.reshape(np.size(b),1)
            
            # use least squares only to get sign of factor
            # get scale factor just by matching rms values
            LS = np.linalg.lstsq(a2d.real, b2d.real, rcond=None)
            sign = np.sign(LS[0][0])
            f += [sign*np.sqrt(np.mean(np.abs(b)**2)) / np.sqrt(np.mean(np.abs(a)**2))]
            
            
        f = np.array(f)
        factors.append(f)
    
    return factors



def calculate_cross_spectrum_matrix(time_data, time_range=None, window=None, N_frames=1, overlap=0.5):
    '''
    Args:
        time_data (<TimeData> object): time series data
        time_range: 2x1 numpy array to specify data segment to use
        window (None or str): apply filter to data before fft or not
        N_frames (int): number of frames to average over
        overlap (between 0,1): frame overlap fraction
    '''
    # TODO iterate over list of timedata... but need new dataset type?
    
    if window==None:
        window='boxcar'
        
    if time_data.__class__.__name__ != 'TimeData':
        raise Exception('Input data needs to be single <TimeData> object')

    if time_range == None:
        ### use all data
        time_range = time_data.time_axis[[0,-1]]
        
    elif time_range.__class__.__name__ == 'PlotData':
        time_range=time_range.ax.get_xbound()
        
    settings = copy.copy(time_data.settings)
    settings.window = window
    settings.time_range = time_range
    settings.N_frames = N_frames
    settings.overlap = overlap

    ## Select data range to use
    s1 = time_data.time_axis >= time_range[0]
    s2 = time_data.time_axis <= time_range[1]
    selection = s1 & s2
    data_selected = time_data.time_data[selection,:]
    #time_selected = timedata.time_axis[selection]
    
    N_samples = len(data_selected[:,0])
    nperseg = np.int32(np.ceil(N_samples / (N_frames+1) / (1-overlap)))
    freqlength = len(np.fft.rfftfreq(nperseg))
    

    noverlap = np.ceil(overlap*nperseg)
    
    Pxy = np.zeros([settings.channels,settings.channels,freqlength],dtype=complex)
    Cxy = np.zeros([settings.channels,settings.channels,freqlength])
    for nx in np.arange(settings.channels):
        for ny in np.arange(settings.channels):
            if nx > ny:
                Pxy[nx,ny,:] = np.conjugate(Pxy[ny,nx,:])
                Cxy[nx,ny,:] = Cxy[ny,nx,:]
            else:
                x = data_selected[:,nx]
                y = data_selected[:,ny]
                f,P = signal.csd(x,y,settings.fs,window=window, nperseg=nperseg, noverlap=noverlap,scaling='spectrum')
                f,C = signal.coherence(x,y,settings.fs,window=window, nperseg=nperseg, noverlap=noverlap)
                Pxy[nx,ny,:] = P
                Cxy[nx,ny,:] = C
            
    cross_spec_data = datastructure.CrossSpecData(f,Pxy,Cxy,settings,id_link=time_data.unique_id,test_name=time_data.test_name)
    
    return cross_spec_data


def calculate_cross_spectra_averaged(time_data_list, time_range=None, window=None):
    '''
    Calculates cross spectra averaged across ensemble of time_data_list. Note that 
    this expects a <TimeDataList> of <TimeData> objects.
    
    Takes each time series as an independent measurement.
    
    Intended for averaged transfer functions from separate measurements, e.g. impulse hammer tests.
    
    Does not average data across sub-frames.    
    
    Args:
        time_data_list (<TimeDataList> object): a list of time series data
        time_range: 2x1 numpy array to specify data segment to use
        window (None or str): type of window to use, default is None.
    '''
    
    if time_data_list.__class__.__name__ != 'TimeDataList':
        raise Exception('Input argument must be <TimeDataList> object.')

    id_link_list = []
    for td in time_data_list:
        id_link_list += [td.unique_id]
    
    settings = copy.copy(time_data_list[0].settings)
    settings.window = window
    settings.time_range = time_range

    
    N_ensemble = len(time_data_list)
    Pxy_av = 0
    for td in time_data_list:
        cross_spec_data = calculate_cross_spectrum_matrix(td, time_range=time_range, window=window, N_frames=1)
        Pxy_av += cross_spec_data.Pxy / N_ensemble
    
    ch_all = np.arange(time_data_list[0].settings.channels)
    Cxy = np.zeros([len(ch_all),len(ch_all),len(Pxy_av[0,0,:])])
    for ch_in in ch_all:
        for ch_out in ch_all:
            Cxy[ch_in,ch_out,:] = np.abs(Pxy_av[ch_in,ch_out,:])**2 / (np.abs(Pxy_av[ch_in,ch_in,:]) * np.abs(Pxy_av[ch_out,ch_out,:]))
    
    
    cross_spec_data_av = datastructure.CrossSpecData(cross_spec_data.freq_axis,Pxy_av,Cxy,settings,id_link=id_link_list,test_name=time_data_list[0].test_name)

    return cross_spec_data_av


def calculate_tf(time_data, ch_in=0, time_range=None, window=None, N_frames=1, overlap=0.5):
    '''
    Args:
        time_data (<TimeData> object): time series data
        ch_in (int): index of input channel
        time_range: 2x1 numpy array to specify data segment to use
        window (None or str): apply filter to data before fft or not
        N_frames (int): number of frames to average over
        overlap (between 0,1): frame overlap fraction
    '''
    if time_data.__class__.__name__ != 'TimeData':
        raise Exception('Input data needs to be single <TimeData> object')


    settings = copy.copy(time_data.settings)
    settings.window = window
    settings.time_range = time_range
    
    ## compute cross spectra
    cross_spec_data = calculate_cross_spectrum_matrix(time_data, time_range=time_range, window=window, N_frames=N_frames, overlap=overlap)
    f = cross_spec_data.freq_axis
    Pxy = cross_spec_data.Pxy
    Cxy = cross_spec_data.Cxy
    ## identify transfer functions and corresponding coherence
    
    ch_all = np.arange(time_data.settings.channels)
    ch_out_set = np.setxor1d(ch_all,ch_in)
    
    tf_data = np.zeros([len(f),len(ch_out_set)],dtype=complex)
    tf_coherence = np.zeros([len(f),len(ch_out_set)])
    ch_count = -1
    for ch_out in ch_out_set:
        ch_count += 1
        tf_data[:,ch_count] = Pxy[ch_in,ch_out,:] / Pxy[ch_in,ch_in,:]
        tf_coherence[:,ch_count] = Cxy[ch_out,ch_in,:]
        
    settings.ch_in = ch_in
    settings.ch_out_set = ch_out_set
    
    tfdata = datastructure.TfData(f,tf_data,tf_coherence,settings,id_link=time_data.unique_id,test_name=time_data.test_name)
    
    return tfdata
    

def calculate_tf_averaged(time_data_list, ch_in=0, time_range=None, window=None):
    '''
    Calculates transfer function averaged across ensemble of timedata. Note that 
    this expects a Python list of timedata objects.
    
    Takes each time series as an independent measurement.
    
    Intended for averaged transfer functions from separate measurements, e.g. impulse hammer tests.
    
    Does not average data across sub-frames.    
    
    Args:
        time_data_list (<TimeDataList> object): a list of time series data
        ch_in (int): index of input channel
        time_range: 2x1 numpy array to specify data segment to use
        window (None or str): type of window to use, default is None.
    '''
    
    if time_data_list.__class__.__name__ != 'TimeDataList':
        raise Exception('Input argument must be <TimeDataList> object.')


    id_link_list = []
    for td in time_data_list:
        id_link_list += [td.unique_id]
        
    N_ensemble = len(time_data_list)
    Pxy_av = 0
    count = -1
    for td in time_data_list:
        count += 1
        ch_all = np.arange(td.settings.channels)
        ch_out_set = np.setxor1d(ch_all,ch_in)
        cross_spec_data = calculate_cross_spectrum_matrix(td, time_range=time_range, window=window, N_frames=1)
        f = cross_spec_data.freq_axis
        Pxy = cross_spec_data.Pxy
        Pxy_av += Pxy / N_ensemble
    
    tf_data = np.zeros([len(f),len(ch_out_set)],dtype=complex)
    tf_coherence = np.zeros([len(f),len(ch_out_set)])    
    ch_count = -1
    for ch_out in ch_out_set:
        ch_count += 1
        tf_data[:,ch_count] = Pxy_av[ch_out,ch_in,:] / Pxy_av[ch_in,ch_in,:]
        tf_coherence[:,ch_count] = np.abs(Pxy_av[ch_in,ch_out,:])**2 / (np.abs(Pxy_av[ch_in,ch_in,:]) * np.abs(Pxy_av[ch_out,ch_out,:]))

    settings = copy.copy(td.settings)
    settings.window = window
    settings.time_range = time_range
    settings.ch_in = ch_in
    settings.ch_out_set = ch_out_set
    
    
    tfdata = datastructure.TfData(f,tf_data,tf_coherence,settings,id_link=id_link_list,test_name=time_data_list[0].test_name)
    
    return tfdata


#%% CLEAN IMPULSE
def clean_impulse(time_data, ch_impulse=0):
    '''
    Sets all data outside of impulse to zero.
    
    Pulse width is estimated by assuming half cosine impulse, using width of half peak amplitude.
    
    Data before peak is unchanged. Data after estimated end of impulse is ramped to zero using half cosine pulse of width 10x estimated pulse width.
    '''
    global MESSAGE
    if not hasattr(time_data,'impulse_cleaned'):
        time_data.impulse_cleaned = False
    
    if time_data.impulse_cleaned == False:    
        y = copy.deepcopy(time_data.time_data[:,ch_impulse])
        yi_max = np.argmax(np.abs(y))
        y_max = np.max(np.abs(y))
        yi_out = np.where(np.abs(y)<y_max/2)[0]
        yi_out1 = yi_out[yi_out < yi_max]
        yi_out2 = yi_out[yi_out > yi_max]
        y1 = yi_out1[-1]
        y2 = yi_out2[0]
        
        
        N = y2-y1
        b = np.int(3*N/2) #half cosine estimate
        end = np.int(yi_max + b/2)
        b = 10*b # less agressive roll off
        #print(time_data.settings.fs/b)
    
        ramp = np.hanning(2*b+1)    
        win = np.ones(len(y))
        win[end:end+b+1] = ramp[b:2*b+1]
        win[end+b:] = 0
        
        y2 = win * y
        
        td = copy.deepcopy(time_data)
        td.time_data[:,ch_impulse] = y2
        td.impulse_cleaned = True
        
        yd = y2-y
        if np.max(np.abs(yd)) > 0.1*np.max(np.abs(y)):
            MESSAGE = 'Cleaned impulse data contained significant signal content.\n'
            MESSAGE += 'Check for possible multiple impacts, or correct channel using ch_impulse.'
        else:
            MESSAGE = 'Impulse data cleaned.'
        print(MESSAGE)
        
        return td
    
    else:
        print('Impulse data already cleaned. Returning copy of input data.')
        return time_data

    
    
#%% SONOGRAM    
def calculate_sonogram(time_data, nperseg=None):
    
    y = np.copy(time_data.time_data) # handles all channels simultaneously
    if nperseg == None:
        nperseg = np.int(len(time_data.time_axis)/50) #roughly 50 fft's per time-series not counting overlap
    f,t,S = signal.spectrogram(y,fs=time_data.settings.fs,nperseg=nperseg,axis=0,mode='complex')
    
    # put channel axis at end
    S_all_chans = np.swapaxes(S,1,2)
    
    sono_data = datastructure.SonoData(t,f,S_all_chans,time_data.settings,id_link=time_data.id_link,test_name=time_data.test_name)
    
    return sono_data

#%% CWT
#def calculate_cwt(time_data):
#    return None


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 27 14:32:35 2018

@author: tb267
"""

import os.path
import numpy as np
import scipy.io as io

from pyqtgraph.Qt import QtGui, QtWidgets


def load_data(filename=None):
    '''
    Loads dataset from filename, or displays a dialog if no argument provided.
    '''
    if filename is None:
        wid = QtWidgets.QWidget()
        filename, _ = QtGui.QFileDialog.getOpenFileName(wid, 'Open data file', '', '*.npy')
        if not filename:
            return None

    d = np.load(filename,allow_pickle=True)
    dataset = d[0]
    return dataset


def save_data(dataset, filename=None, overwrite_without_prompt=False):
    '''
    Saves dataset class to file 'filename.npy', or provides dialog if no
    filename provided.

    Args:
       dataset: An object of the class dataSet
       filename: string [optional]
       overwrite_without_prompt: bool

    '''

    # put data into numpy array
    d = np.array([dataset])

    # If filename not specified, provide dialog
    if filename is None:
        wid = QtWidgets.QWidget()
        filename, _ = QtGui.QFileDialog.getSaveFileName(wid, 'Save dataset', '', '*.npy')
        if not filename:
            # No filename chosen, give up on saving
            print('Save cancelled')
            return None


    # If it exists, check if we should overwrite it (unless
    # overwrite_without_prompt is True)
    elif os.path.isfile(filename) and not overwrite_without_prompt:
        answer = input('File %r already exists. Overwrite? [y/n]: ' % filename)
        if answer != 'y':
            print('Save cancelled')
            return None
        print('Will overwrite existing file')
        
    # Make sure it ends with .npy
    if not filename.endswith('.npy'):
        filename += '.npy'
        
    # Actually save!
    np.save(filename, d)
    print("Data saved as %s" % filename)

    return filename



def save_fig(plot, figsize=None, filename=None, overwrite_without_prompt=False):
    '''
    Saves figure to file 'filename.png' and 'filename.pdf', or provides dialog if no
    filename provided.

    Args:
       fig: A matplotlib fig object
       filename: string [optional]
       overwrite_without_prompt: boo
    '''
    if plot.__class__.__name__ is 'PlotData':
        fig = plot.fig
    elif plot.__class__.__name__ is 'Figure':
        fig = plot

    # If filename not specified, provide dialog
    if filename is None:
        wid = QtWidgets.QWidget()
        filename, _ = QtGui.QFileDialog.getSaveFileName(wid, 'Save figure', '')
        if not filename:
            # No filename chosen, give up on saving
            print('Save cancelled')
            return None


    # If it exists, check if we should overwrite it (unless
    # overwrite_without_prompt is True)
    elif os.path.isfile(filename) and not overwrite_without_prompt:
        answer = input('File %r already exists. Overwrite? [y/n]: ' % filename)
        if answer != 'y':
            print('Save cancelled')
            return None
        print('Will overwrite existing file')

    # Set figsize...
    original_size = fig.get_size_inches()
    if figsize is not None:
        fig.set_size_inches(figsize,forward=False)
        
    # Make sure it ends with .png then .pdf
    filename = os.path.splitext(filename)[0]
    if not filename.endswith('.png'):
        filename += '.png'
    fig.savefig(filename, dpi=300)
    print("Figure saved as %s" % filename)
    
    filename = os.path.splitext(filename)[0]
    if not filename.endswith('.pdf'):
        filename += '.pdf'
    fig.savefig(filename, dpi=300)
    print("Figure saved as %s" % filename)

    # return to original size
    fig.set_size_inches(original_size,forward=False)
    
    
    return filename



#%% EXPORT TO MATLAB
def export_to_matlab(dataset, filename=None, overwrite_without_prompt=False):
    '''
    Exports dataset class to file 'filename.mat', or provides dialog if no
    filename provided.
    
    Saved file can be loaded directly in Matlab as set of arrays.

    Args:
       dataset: An object of the class dataSet
       filename: string [optional]
       overwrite_without_prompt: bool

    '''
    
    # convert data into dictionary ready for Matlab
    data_matlab = dict()
    
    #%% TIME
    if len(dataset.time_data_list) > 0:
        T=0
        fs=0
        n_time=0
        for time_data in dataset.time_data_list:
            N = len(time_data.time_axis)
            T = np.max([time_data.time_axis[-1]*N/(N-1),T])
            fs = np.max([1/np.mean(np.diff(time_data.time_axis)),fs])
            n_time += time_data.settings.channels
        
        t=np.arange(0,T,1/fs)
        time_data_all = np.zeros((len(t),n_time))
        counter = -1
        for time_data in dataset.time_data_list:
            for i in range(time_data.settings.channels):
                counter += 1
                time_data_all[:,counter] = np.interp(t,time_data.time_axis,time_data.time_data[:,i],right=0)
                
        data_matlab['time_axis_all'] = t
        data_matlab['time_data_all'] = time_data_all

    


    #%% FFT - doesn't export coherence
    if len(dataset.freq_data_list) > 0:
        df=np.inf
        fmax=0
        n_tf=0
        for freq_data in dataset.freq_data_list:
            df_check = np.mean(np.diff(freq_data.freq_axis))
            df = np.min([df,df_check])
            fmax = np.max([freq_data.freq_axis[-1],fmax])
            tf_shape = np.shape(freq_data.freq_data)
            n_tf += tf_shape[1]
        
        f=np.arange(0,fmax+df,df)
        npts = 2*(len(f)-1)
        fs_tf = 2*f[-1]
        freq_data_all = np.zeros((len(f),n_tf),dtype=complex)
        counter = -1
        for freq_data in dataset.freq_data_list:
            freq_shape = np.shape(freq_data.freq_data)
            for i in range(freq_shape[1]):
                counter += 1
                freq_data_all[:,counter] = np.interp(f,freq_data.freq_axis,freq_data.freq_data[:,i],right=0)
        
        data_matlab['freq_axis_all'] = f
        data_matlab['freq_data_all'] = freq_data_all
        
 


    #%% Transfer Function - doesn't export coherence
    if len(dataset.tf_data_list) > 0:
        df=np.inf
        fmax=0
        n_tf=0
        for tf_data in dataset.tf_data_list:
            df_check = np.mean(np.diff(tf_data.freq_axis))
            df = np.min([df,df_check])
            fmax = np.max([tf_data.freq_axis[-1],fmax])
            tf_shape = np.shape(tf_data.tf_data)
            n_tf += tf_shape[1]
        
        f=np.arange(0,fmax+df,df)
        npts = 2*(len(f)-1)
        fs_tf = 2*f[-1]
        tf_data_all = np.zeros((len(f),n_tf),dtype=complex)
        counter = -1
        for tf_data in dataset.tf_data_list:
            tf_shape = np.shape(tf_data.tf_data)
            for i in range(tf_shape[1]):
                counter += 1
                tf_data_all[:,counter] = np.interp(f,tf_data.freq_axis,tf_data.tf_data[:,i],right=0)
        
        data_matlab['tf_axis_all'] = f
        data_matlab['tf_data_all'] = tf_data_all
        

    


    #%% SAVE

    # If filename not specified, provide dialog
    if filename is None:
        wid = QtWidgets.QWidget()
        filename, _ = QtGui.QFileDialog.getSaveFileName(wid, 'Save dataset', '', '*.mat')
        if not filename:
            # No filename chosen, give up on saving
            print('Save cancelled')
            return None


    # If it exists, check if we should overwrite it (unless
    # overwrite_without_prompt is True)
    elif os.path.isfile(filename) and not overwrite_without_prompt:
        answer = input('File %r already exists. Overwrite? [y/n]: ' % filename)
        if answer != 'y':
            print('Save cancelled')
            return None
        print('Will overwrite existing file')
        
    # Make sure it ends with .npy
    if not filename.endswith('.mat'):
        filename += '.mat'
        
    # Actually save!
    io.savemat(filename,data_matlab)
    print("Data saved as %s" % filename)

    return filename



#%% EXPORT TO MATLAB JWLOGGER
def export_to_matlab_jwlogger(dataset, filename=None, overwrite_without_prompt=False):
    '''
    Exports dataset class to file 'filename.mat', or provides dialog if no
    filename provided.
    
    Saved file is compatible with Jim Woodhouse logger file format.

    Args:
       dataset: An object of the class dataSet
       filename: string [optional]
       overwrite_without_prompt: bool

    '''

    # convert data into dictionary ready for Matlab
    data_jwlogger = dict()
    
    #%% TIME
    if len(dataset.time_data_list) > 0:
        T=0
        fs=0
        n_time=0
        for time_data in dataset.time_data_list:
            N = len(time_data.time_axis)
            T = np.max([time_data.time_axis[-1]*N/(N-1),T])
            fs = np.max([1/np.mean(np.diff(time_data.time_axis)),fs])
            n_time += time_data.settings.channels
        
        t=np.arange(0,T,1/fs)
        time_data_all = np.zeros((len(t),n_time))
        counter = -1
        for time_data in dataset.time_data_list:
            for i in range(time_data.settings.channels):
                counter += 1
                time_data_all[:,counter] = np.interp(t,time_data.time_axis,time_data.time_data[:,i],right=0)
                
        data_jwlogger['buflen'] = np.float(np.size(t))
        data_jwlogger['indata'] = time_data_all
        data_jwlogger['tsmax'] = np.float(t[-1])
    else:
        n_time = 0
        time_data_all = 0
    
    
    #%% Transfer Function - doesn't export coherence
    if len(dataset.tf_data_list) > 0:
        df=np.inf
        fmax=0
        n_tf=0
        for tf_data in dataset.tf_data_list:
            df_check = np.mean(np.diff(tf_data.freq_axis))
            df = np.min([df,df_check])
            fmax = np.max([tf_data.freq_axis[-1],fmax])
            tf_shape = np.shape(tf_data.tf_data)
            n_tf += tf_shape[1]
        
        f=np.arange(0,fmax+df,df)
        npts = 2*(len(f)-1)
        fs_tf = 2*f[-1]
        tf_data_all = np.zeros((len(f),n_tf),dtype=complex)
        counter = -1
        for tf_data in dataset.tf_data_list:
            tf_shape = np.shape(tf_data.tf_data)
            for i in range(tf_shape[1]):
                counter += 1
                tf_data_all[:,counter] = np.interp(f,tf_data.freq_axis,tf_data.tf_data[:,i],right=0)
        
        # convert
        data_jwlogger['freq'] = np.float(fs_tf)
        data_jwlogger['npts'] = np.float(npts)
        data_jwlogger['yspec'] = tf_data_all
    else:
        n_tf = 0
        tf_data_all = 0
    
    #%% Convert
    
    data_jwlogger['dt2'] = np.array([n_time,n_tf,0],dtype=float)
    data_jwlogger['dtype'] = np.array([n_time,n_tf,0],dtype=float)
    

    # SAVE

    # If filename not specified, provide dialog
    if filename is None:
        wid = QtWidgets.QWidget()
        filename, _ = QtGui.QFileDialog.getSaveFileName(wid, 'Save dataset', '', '*.mat')
        if not filename:
            # No filename chosen, give up on saving
            print('Save cancelled')
            return None


    # If it exists, check if we should overwrite it (unless
    # overwrite_without_prompt is True)
    elif os.path.isfile(filename) and not overwrite_without_prompt:
        answer = input('File %r already exists. Overwrite? [y/n]: ' % filename)
        if answer != 'y':
            print('Save cancelled')
            return None
        print('Will overwrite existing file')
        
    # Make sure it ends with .npy
    if not filename.endswith('.mat'):
        filename += '.mat'
        
    # Actually save!
    io.savemat(filename,data_jwlogger)
    print("Data saved as %s" % filename)

    return filename


def export_to_csv(data_list, filename=None, overwrite_without_prompt=False):
    '''
    Exports data to file 'filename.csv', or provides dialog if no
    filename provided.
    
    Saved file is *.csv

    Args:
       data_list: An object of the class TimeDataList, FreqDataList, or TfDataList
       filename: string [optional]
       overwrite_without_prompt: bool
    '''
    
    data_list_type = data_list.__class__.__name__
    
    if data_list_type == 'TimeDataList':
        darray = np.transpose(np.atleast_2d(data_list[0].time_axis))
        for time_data in data_list:
            darray = np.append(darray,time_data.time_data,axis=1)
        
            
            
    elif data_list_type == 'FreqDataList':
        darray = np.transpose(np.atleast_2d(data_list[0].freq_axis))
        for freq_data in data_list:
            darray = np.append(darray,freq_data.freq_data,axis=1)
        
    elif data_list_type == 'TfDataList':
        darray = np.transpose(np.atleast_2d(data_list[0].freq_axis))
        for tf_data in data_list:
            darray = np.append(darray,tf_data.tf_data,axis=1)
        
    else:
        print('Expecting input to be one of TimeDataList, FreqDataList, or TfDataList')
        return None
    
    # SAVE

    # If filename not specified, provide dialog
    if filename is None:
        wid = QtWidgets.QWidget()
        filename, _ = QtGui.QFileDialog.getSaveFileName(wid, 'Save dataset', '', '*.csv')
        if not filename:
            # No filename chosen, give up on saving
            print('Save cancelled')
            return None


    # If it exists, check if we should overwrite it (unless
    # overwrite_without_prompt is True)
    elif os.path.isfile(filename) and not overwrite_without_prompt:
        answer = input('File %r already exists. Overwrite? [y/n]: ' % filename)
        if answer != 'y':
            print('Save cancelled')
            return None
        print('Will overwrite existing file')
        
    # Make sure it ends with .csv
    if not filename.endswith('.csv'):
        filename += '.csv'
        
    # Actually save!
    np.savetxt(filename, darray, delimiter=",")
    print("Data saved as %s" % filename)

    return filename


from . import plotting
from . import datastructure
from . import acquisition
from . import analysis
from . import streams
from . import file
import time

import numpy as np
from IPython.display import display
from ipywidgets import Button, HBox, VBox, Output, FloatText, IntText, Dropdown, IntSlider, Layout, Label, Checkbox,ToggleButton


#%%
class InteractiveLogging():
    def __init__(self,settings,test_name=None,default_window='hanning'):
        
        if default_window is None:
            default_window = 'None'
        self.settings = settings
        self.test_name = test_name
        self.dataset = datastructure.DataSet()
        
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out = Output()
        self.out_top = Output(layout=Layout(height='120px'))
        
        
        # Initialise variables
        self.current_view = 'Time'    
        self.N_frames = 1
        self.overlap = 0.5
        self.iw_fft_power = 0
        self.iw_tf_power = 0
        self.legend_loc = 'lower right'

        # puts plot inside widget so can have buttons next to plot
        self.outplot = Output(layout=Layout(width='85%'))
                
        # BUTTONS
        items_measure = ['Log Data','Delete Last Measurement','Reset (clears all data)','Load Data']
        items_view = ['View Time', 'View FFT', 'View TF']
        items_calc = ['Calc FFT', 'Calc TF', 'Calc TF average']
        items_axes = ['xmin','xmax','ymin','ymax']
        items_save = ['Save Dataset','Save Figure']
        items_iw   = ['multiply iw','divide iw']
        
        items_legend = ['Left','on/off','Right']
        
        
        
        self.buttons_measure= [Button(description=i, layout=Layout(width='33%')) for i in items_measure]
        self.buttons_measure[0].button_style ='success'
        self.buttons_measure[0].style.font_weight = 'bold'
        self.buttons_measure[1].button_style ='warning'
        self.buttons_measure[1].style.font_weight = 'bold'
        self.buttons_measure[2].button_style ='danger'
        self.buttons_measure[2].style.font_weight = 'bold'
        self.buttons_measure[3].button_style ='primary'
        self.buttons_measure[3].style.font_weight = 'bold'
        
        self.buttons_view = [Button(description=i, layout=Layout(width='95%')) for i in items_view]
        self.buttons_view[0].button_style ='info'
        self.buttons_view[1].button_style ='info'
        self.buttons_view[2].button_style ='info'
        
        self.buttons_calc = [Button(description=i, layout=Layout(width='99%')) for i in items_calc]
        self.buttons_calc[0].button_style ='primary'
        self.buttons_calc[1].button_style ='primary'
        self.buttons_calc[2].button_style ='primary'

        
        self.buttons_iw_fft = [Button(description=i, layout=Layout(width='50%')) for i in items_iw]
        self.buttons_iw_fft[0].button_style = 'info'
        self.buttons_iw_fft[1].button_style = 'info'
        
        self.buttons_iw_tf = [Button(description=i, layout=Layout(width='50%')) for i in items_iw]
        self.buttons_iw_tf[0].button_style = 'info'
        self.buttons_iw_tf[1].button_style = 'info'
        
        self.buttons_match = Button(description='Match Amplitudes', layout=Layout(width='99%'))
        self.buttons_match.button_style = 'info'
        
        self.buttons_save = [Button(description=i, layout=Layout(width='50%')) for i in items_save]
        self.buttons_save[0].button_style = 'success'
        self.buttons_save[0].style.font_weight = 'bold'
        self.buttons_save[1].button_style = 'success'
        self.buttons_save[1].style.font_weight = 'bold'
        
        self.button_warning = Button(description='WARNING: Data may be clipped. Press here to delete last measurement.', layout=Layout(width='100%'))
        self.button_warning.button_style = 'danger'
        self.button_warning.style.font_weight='bold'
        
        self.button_X = Button(description='Auto X', layout=Layout(width='95%'))
        self.button_Y = Button(description='Auto Y', layout=Layout(width='95%'))
        self.button_X.button_style ='success'
        self.button_Y.button_style ='success'
        
        self.buttons_legend = [Button(description=i, layout=Layout(width='31%')) for i in items_legend]
        self.buttons_legend_toggle = ToggleButton(description='Click-and-drag', layout=Layout(width='95%',alignitems='start'))
        
        # TEXT/LABELS/DROPDOWNS
        self.item_iw_fft_label = Label(value='iw power={}'.format(0),layout=Layout(width='100%'))
        self.item_iw_tf_label  = Label(value='iw power={}'.format(0),layout=Layout(width='100%'))
        self.item_label = Label(value="Frame length = {:.2f} seconds.".format(settings.stored_time/(self.N_frames-self.overlap*self.N_frames+self.overlap)))
        self.item_axis_label = Label(value="Axes control:",layout=Layout(width='95%'))
        self.item_view_label = Label(value="View data:",layout=Layout(width='95%'))
        self.item_legend_label = Label(value="Legend position:",layout=Layout(width='95%'))
        self.item_blank_label = Label(value="",layout=Layout(width='95%'))
        
        self.text_axes = [FloatText(value=0,description=i, layout=Layout(width='95%')) for i in items_axes]
        self.text_axes =  [self.button_X]+[self.button_Y] + self.text_axes
        self.drop_window = Dropdown(options=['None', 'hanning'],value=default_window,description='Window:', layout=Layout(width='99%'))
        self.slide_Nframes = IntSlider(value=1,min=1,max=30,step=1,description='N_frames:',continuous_update=True,readout=False, layout=Layout(width='99%'))
        self.text_Nframes = IntText(value=1,description='N_frames:', layout=Layout(width='99%'))
        
        # VERTICAL GROUPS
        
        group0 = VBox([self.buttons_calc[0],self.drop_window,HBox(self.buttons_iw_fft)],layout=Layout(width='33%'))
        group1 = VBox([self.buttons_calc[1],self.drop_window,self.slide_Nframes,self.text_Nframes,self.item_label,HBox(self.buttons_iw_tf),self.buttons_match],layout=Layout(width='33%'))
        group2 = VBox([self.buttons_calc[2],self.drop_window,HBox(self.buttons_iw_tf),self.buttons_match],layout=Layout(width='33%'))
        
        group_view = VBox([self.item_axis_label]+self.text_axes+[self.item_legend_label,self.buttons_legend_toggle]+[HBox(self.buttons_legend),self.item_view_label]+self.buttons_view,layout=Layout(width='20%'))
        
        # ASSEMBLE
        display(self.out_top)
        display(HBox([self.button_warning]))
        display(HBox(self.buttons_measure))
        display(HBox([self.outplot,group_view]))
        display(HBox([group0,group1,group2]))
        display(HBox(self.buttons_save))
        self.button_warning.layout.visibility='hidden'
        
        # second part to putting plot inside widget
        with self.outplot:
            self.p = plotting.PlotData(figsize=(7.5,4))
        
        
        ## Make buttons/boxes interactive
        
        self.text_axes[2].observe(self.xmin, names='value')
        self.text_axes[3].observe(self.xmax, names='value')
        self.text_axes[4].observe(self.ymin, names='value')
        self.text_axes[5].observe(self.ymax, names='value')
        
        self.button_X.on_click(self.auto_x)
        self.button_Y.on_click(self.auto_y)
        
        self.buttons_legend[0].on_click(self.legend_left)
        self.buttons_legend[1].on_click(self.legend_onoff)
        self.buttons_legend[2].on_click(self.legend_right)
        self.buttons_legend_toggle.observe(self.legend_toggle)
        
        self.slide_Nframes.observe(self.nframes_slide)
        self.text_Nframes.observe(self.nframes_text)
        
        self.buttons_measure[0].on_click(self.measure)
        self.buttons_measure[1].on_click(self.undo)
        self.buttons_measure[2].on_click(self.reset)
        self.buttons_measure[3].on_click(self.load_data)
        
        self.buttons_view[0].on_click(self.view_time)
        self.buttons_view[1].on_click(self.view_fft)
        self.buttons_view[2].on_click(self.view_tf)
        
        self.buttons_calc[0].on_click(self.fft)
        self.buttons_calc[1].on_click(self.tf)
        self.buttons_calc[2].on_click(self.tf_av)
        
        self.buttons_iw_fft[0].on_click(self.xiw_fft)
        self.buttons_iw_fft[1].on_click(self.diw_fft)
        self.buttons_iw_tf[0].on_click(self.xiw_tf)
        self.buttons_iw_tf[1].on_click(self.diw_tf)
        
        self.buttons_match.on_click(self.match)
        
        self.buttons_save[0].on_click(self.save_data)
        self.buttons_save[1].on_click(self.save_fig)
        
        self.button_warning.on_click(self.undo)
        
        self.refresh_buttons()
        
        
        # Put output text at bottom of display
        display(self.out)
        
        with self.out_top:
            try:
                streams.start_stream(settings)
                self.rec = streams.REC
            except:
                print('Data stream not initialised.')
                print('Possible reasons: pyaudio or PyDAQmx not installed, or acquisition hardware not connected.')
                print('Please note that it won''t be possible to log data.')
            
        
    
    def xmin(self,v):
        xmin = self.text_axes[2].value
        xlim = self.p.ax.get_xlim()
        self.p.ax.set_xlim([xmin,xlim[1]])
        
    def xmax(self,v):
        xmax = self.text_axes[3].value
        xlim = self.p.ax.get_xlim()
        self.p.ax.set_xlim([xlim[0],xmax])
        
    def ymin(self,v):
        ymin = self.text_axes[4].value
        ylim = self.p.ax.get_ylim()
        self.p.ax.set_ylim([ymin,ylim[1]])
        
    def ymax(self,v):
        ymax = self.text_axes[5].value
        ylim = self.p.ax.get_ylim()
        self.p.ax.set_ylim([ylim[0],ymax])
        
    def auto_x(self,b):
        self.p.auto_x()
        
    def auto_y(self,b):
        self.p.auto_y()
        
    def legend_left(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.legend_loc = 'lower left'
            self.p.update_legend(self.legend_loc)
        
    def legend_right(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.legend_loc = 'lower right'
            self.p.update_legend(self.legend_loc)
            
    def legend_onoff(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            visibility = self.p.ax.get_legend().get_visible()
            self.p.ax.get_legend().set_visible(not visibility)
                
    def legend_toggle(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.p.ax.get_legend().set_visible(True)
            self.p.legend.set_draggable(self.buttons_legend_toggle.value)
            
        
    def nframes_slide(self,v):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.N_frames = self.slide_Nframes.value
            self.text_Nframes.value = self.N_frames
            if len(self.dataset.time_data_list) is not 0:
                stored_time = self.dataset.time_data_list[0].settings.stored_time
            elif len(self.dataset.tf_data_list) is not 0:
                stored_time = self.dataset.tf_data_list[0].settings.stored_time
            else:
                stored_time = 0
                print('Time or TF data settings not found')
                
            self.item_label.value = "Frame length = {:.2f} seconds.".format(stored_time/(self.N_frames-self.overlap*self.N_frames+self.overlap))
            if self.current_view is 'TF':
                self.tf(None)
        
    def nframes_text(self,v):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.N_frames = self.text_Nframes.value
            self.slide_Nframes.value = self.N_frames
            if len(self.dataset.time_data_list) is not 0:
                stored_time = self.dataset.time_data_list[0].settings.stored_time
            elif len(self.dataset.tf_data_list) is not 0:
                stored_time = self.dataset.tf_data_list[0].settings.stored_time
            else:
                stored_time = 0
                print('Time or TF data settings not found')
            self.item_label.value = "Frame length = {:.2f} seconds.".format(stored_time/(self.N_frames-self.overlap*self.N_frames+self.overlap))
            if self.current_view is 'TF':
                self.tf(None)
        
    def measure(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.rec.trigger_detected = False
            self.buttons_measure[0].button_style =''
            if self.settings.pretrig_samples is None:
                self.buttons_measure[0].description = 'Logging ({}s)'.format(self.settings.stored_time)
            else:
                self.buttons_measure[0].description = 'Logging ({}s, with trigger)'.format(self.settings.stored_time)
            
            d = acquisition.log_data(self.settings,test_name=self.test_name, rec=self.rec)
            self.dataset.add_to_dataset(d.time_data_list)
            N = len(self.dataset.time_data_list)
            self.p.update(self.dataset.time_data_list,sets=[N-1],channels='all')
            self.p.auto_x()
#            self.p.auto_y()
            self.p.ax.set_ylim([-1,1])
            self.current_view='Time'
            self.buttons_measure[0].button_style ='success'
            self.buttons_measure[0].description = 'Log Data'
            
            if np.any(np.abs(d.time_data_list[-1].time_data) > 0.95):
                self.button_warning.layout.visibility = 'visible'
            else:
                self.button_warning.layout.visibility = 'hidden'
            
            xlim = self.p.ax.get_xlim()
            ylim =  self.p.ax.get_ylim()
            self.text_axes[2].value = xlim[0]
            self.text_axes[3].value = xlim[1]
            self.text_axes[4].value = ylim[0]
            self.text_axes[5].value = ylim[1]
            self.refresh_buttons()
        
    def undo(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.dataset.remove_last_data_item('TimeData')
            self.dataset.freq_data_list = datastructure.FreqDataList()
            self.dataset.tf_data_list = datastructure.TfDataList()
            N = len(self.dataset.time_data_list)
            self.p.update(self.dataset.time_data_list,sets=[N-1],channels='all')
            self.button_warning.layout.visibility = 'hidden'
            self.refresh_buttons()
            
    def reset(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.dataset = datastructure.DataSet()
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            N = len(self.dataset.time_data_list)
            self.p.update(self.dataset.time_data_list,sets=[N-1],channels='all')
            self.refresh_buttons()
        
    
                
    def load_data(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            d = file.load_data()
            if d is not None:
                self.dataset.add_to_dataset(d.time_data_list)
                self.dataset.add_to_dataset(d.freq_data_list)
                self.dataset.add_to_dataset(d.tf_data_list)
                self.dataset.add_to_dataset(d.cross_spec_data_list)
                self.dataset.add_to_dataset(d.sono_data_list)
            else:
                print('No data loaded')
            
            if len(self.dataset.time_data_list) is not 0:
                self.p.update(self.dataset.time_data_list,sets='all',channels='all')
            elif len(self.dataset.freq_data_list) is not 0:
                self.p.update(self.dataset.freq_data_list,sets='all',channels='all')
            elif len(self.dataset.tf_data_list) is not 0:
                self.p.update(self.dataset.tf_data_list,sets='all',channels='all')
            else:
                print('No data to view')
           
            self.refresh_buttons()
        
            self.p.auto_x()
            self.p.auto_y()
            
            xlim = self.p.ax.get_xlim()
            ylim = self.p.ax.get_ylim()
            self.text_axes[2].value = xlim[0]
            self.text_axes[3].value = xlim[1]
            self.text_axes[4].value = ylim[0]
            self.text_axes[5].value = ylim[1]
            
        
        
    def view_time(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.refresh_buttons()
            N = len(self.dataset.time_data_list)
            if N is not 0:
                self.p.update(self.dataset.time_data_list)
                if self.current_view is not 'Time':
                    self.current_view = 'Time'
                    self.p.auto_x()
                    self.p.auto_y()

                xlim = self.p.ax.get_xlim()
                ylim = self.p.ax.get_ylim()
                self.text_axes[2].value = xlim[0]
                self.text_axes[3].value = xlim[1]
                self.text_axes[4].value = ylim[0]
                self.text_axes[5].value = ylim[1]
            else:
                print('no time data to display')
                
    def view_fft(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            N = len(self.dataset.freq_data_list)
            self.refresh_buttons()
            if N is not 0:
                self.p.update(self.dataset.freq_data_list)
                if self.current_view is not 'FFT':
                    self.current_view = 'FFT'
                    self.p.auto_x()
                    self.p.auto_y() 
                    
                xlim = self.p.ax.get_xlim()
                ylim = self.p.ax.get_ylim()
                self.text_axes[2].value = xlim[0]
                self.text_axes[3].value = xlim[1]
                self.text_axes[4].value = ylim[0]
                self.text_axes[5].value = ylim[1]
            else:
                print('no FFT data to display')
                
    def view_tf(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            N = len(self.dataset.tf_data_list)
            self.refresh_buttons()
            if N is not 0:
                self.p.update(self.dataset.tf_data_list)
                if self.current_view is not 'TF':
                    self.current_view = 'TF'
                    self.p.auto_x()
                    self.p.auto_y()
                xlim = self.p.ax.get_xlim()
                ylim = self.p.ax.get_ylim()
                self.text_axes[2].value = xlim[0]
                self.text_axes[3].value = xlim[1]
                self.text_axes[4].value = ylim[0]
                self.text_axes[5].value = ylim[1]
            else:
                print('no TF data to display')
            
    def fft(self,b):
        window = self.drop_window.value
        if window is 'None':
            window = None
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.dataset.calculate_fft_set(window=window)
            self.p.update(self.dataset.freq_data_list)
            if self.current_view is not 'FFT':
                self.current_view = 'FFT'
                self.p.auto_x()
                self.p.auto_y() 
                
            xlim = self.p.ax.get_xlim()
            ylim = self.p.ax.get_ylim()
            self.text_axes[2].value = xlim[0]
            self.text_axes[3].value = xlim[1]
            self.text_axes[4].value = ylim[0]
            self.text_axes[5].value = ylim[1]
            self.refresh_buttons()
        
    
    def tf(self,b):
        N_frames = self.N_frames
        window = self.drop_window.value
        if window is 'None':
            window = None
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.dataset.calculate_tf_set(window=window,N_frames=N_frames,overlap=self.overlap)
            self.p.update(self.dataset.tf_data_list)
            if self.current_view is not 'TF':
                self.current_view = 'TF'
                self.p.auto_x()
                self.p.auto_y()
            xlim = self.p.ax.get_xlim()
            ylim = self.p.ax.get_ylim()
            self.text_axes[2].value = xlim[0]
            self.text_axes[3].value = xlim[1]
            self.text_axes[4].value = ylim[0]
            self.text_axes[5].value = ylim[1]
            self.refresh_buttons()
    
    def tf_av(self,b):
        window = self.drop_window.value
        if window is 'None':
            window = None
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.dataset.calculate_tf_averaged(window=window)
            self.p.update(self.dataset.tf_data_list)
            if self.current_view is not 'TFAV':
                self.current_view = 'TFAV'
                self.p.auto_x()
                self.p.auto_y()
                
            xlim = self.p.ax.get_xlim()
            ylim = self.p.ax.get_ylim()
            self.text_axes[2].value = xlim[0]
            self.text_axes[3].value = xlim[1]
            self.text_axes[4].value = ylim[0]
            self.text_axes[5].value = ylim[1]
            self.refresh_buttons()
            
    def xiw_fft(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            if self.current_view is 'FFT':
                s = self.p.get_selected_channels()
                n_sets,n_chans = np.shape(s)
                for ns in range(n_sets):
                    newdata = analysis.multiply_by_power_of_iw(self.dataset.freq_data_list[ns],power=1,channel_list=s[ns,:])
                    self.dataset.freq_data_list[ns] = newdata
                self.p.update(self.dataset.freq_data_list)
                self.p.auto_y()
                self.iw_fft_power += 1
                print('Multiplied by (iw)**{}'.format(self.iw_fft_power))
            else:
                print('First press <Calc FFT>')
    
    def diw_fft(self,b):
       # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            if self.current_view is 'FFT':
                s = self.p.get_selected_channels()
                n_sets,n_chans = np.shape(s)
                for ns in range(n_sets):
                    newdata = analysis.multiply_by_power_of_iw(self.dataset.freq_data_list[ns],power=-1,channel_list=s[ns,:])
                    self.dataset.freq_data_list[ns] = newdata
                self.p.update(self.dataset.freq_data_list)
                self.p.auto_y()
                self.iw_fft_power -= 1
                print('Multiplied by (iw)**{}'.format(self.iw_fft_power))
            else:
                print('First press <Calc FFT>')
                
    def xiw_tf(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            if (self.current_view is 'TF') or (self.current_view is 'TFAV'):
                s = self.p.get_selected_channels()
                n_sets,n_chans = np.shape(s)
                for ns in range(n_sets):
                    newdata = analysis.multiply_by_power_of_iw(self.dataset.tf_data_list[ns],power=1,channel_list=s[ns,:])
                    self.dataset.tf_data_list[ns] = newdata
                self.p.update(self.dataset.tf_data_list)
                self.p.auto_y()
                self.iw_tf_power += 1
                print('Multiplied selected channel by (iw)**{}'.format(self.iw_tf_power))
            else:
                print('First press <Calc TF> or <Calc TF average>')
                
    def diw_tf(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            if (self.current_view is 'TF') or (self.current_view is 'TFAV'):
                s = self.p.get_selected_channels()
                n_sets,n_chans = np.shape(s)
                for ns in range(n_sets):
                    newdata = analysis.multiply_by_power_of_iw(self.dataset.tf_data_list[ns],power=-1,channel_list=s[ns,:])
                    self.dataset.tf_data_list[ns] = newdata
                self.p.update(self.dataset.tf_data_list)
                self.p.auto_y()
                self.iw_tf_power -= 1
                print('Multiplied selected channel by (iw)**{}'.format(self.iw_tf_power))
            else:
                print('First press <Calc TF> or <Calc TF average>')
                
    def match(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            if (self.current_view is 'TF') or (self.current_view is 'TFAV'):
                freq_range = self.p.ax.get_xlim()
                current_calibration_factors = self.dataset.tf_data_list.get_calibration_factors()
                reference = current_calibration_factors[0][0]
                factors = analysis.best_match(self.dataset.tf_data_list,freq_range=freq_range,set_ref=0,ch_ref=0)
                factors = [reference*x for x in factors]
                self.dataset.tf_data_list.set_calibration_factors_all(factors)
                self.p.update(self.dataset.tf_data_list)
                print('scale factors:')
                print(factors)
                #self.p.auto_y()
            else:
                print('First press <View TF> or <Calc TF> or <Calc TF average>')
    
    def save_data(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            print('Saving dataset:')
            print(self.dataset)
            self.dataset.save_data()
            
    
    def save_fig(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            file.save_fig(self.p,figsize=(9,5))
            
    def refresh_buttons(self):
        if len(self.dataset.time_data_list) is 0:
            self.buttons_view[0].button_style =''
        else:
            self.buttons_view[0].button_style ='info'
            
        if len(self.dataset.freq_data_list) is 0:
            self.buttons_view[1].button_style =''
        else:
            self.buttons_view[1].button_style ='info'
            
        if len(self.dataset.tf_data_list) is 0:
            self.buttons_view[2].button_style =''
        else:
            self.buttons_view[2].button_style ='info'
    




#%%
class InteractiveView():
    def __init__(self):
        
        
        self.dataset = datastructure.DataSet()
        
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out = Output()
        self.out_top = Output()
        
        # Initialise
        self.current_view = 'Time'    
        self.N_frames = 1
        self.overlap = 0.5
        
        # BUTTONS
        items_load = ['Load Data','Reset (deletes loaded data)']
        items_view = ['View Time', 'View FFT', 'View TF']
        items_axes = ['xmin','xmax','ymin','ymax']
        items_save = ['Save Dataset','Save Figure']
        items_legend = ['Left','on/off','Right']
        
        self.buttons_load = [Button(description=i, layout=Layout(width='50%')) for i in items_load]
        self.buttons_load[0].button_style ='primary'
        self.buttons_load[0].style.font_weight = 'bold'
        self.buttons_load[1].button_style ='warning'
        self.buttons_load[1].style.font_weight = 'bold'

        
        self.buttons_view = [Button(description=i, layout=Layout(width='99%')) for i in items_view]
        self.buttons_view[0].button_style ='info'
        self.buttons_view[1].button_style ='info'
        self.buttons_view[2].button_style ='info'
        
        self.buttons_match = Button(description='Match Amplitudes', layout=Layout(width='99%'))
        self.buttons_match.button_style = 'info'
        
        
        self.buttons_save = [Button(description=i, layout=Layout(width='50%')) for i in items_save]
        self.buttons_save[0].button_style = 'success'
        self.buttons_save[0].style.font_weight = 'bold'
        self.buttons_save[1].button_style = 'success'
        self.buttons_save[1].style.font_weight = 'bold'
        
        
        self.button_X = Button(description='Auto X', layout=Layout(width='10%'))
        self.button_Y = Button(description='Auto Y', layout=Layout(width='10%'))
        self.button_X.button_style ='success'
        self.button_Y.button_style ='success'
        
        self.buttons_legend = [Button(description=i, layout=Layout(width='16%')) for i in items_legend]
        self.buttons_legend_toggle = ToggleButton(description='Click-and-drag', layout=Layout(width='16%',alignitems='start'))
        
        # TEXT/LABELS/DROPDOWNS
        self.item_axis_label = Label(value="Adjust axes manually:",layout=Layout(width='16%'))
        self.text_axes = [FloatText(value=0,description=i, layout=Layout(width='16%')) for i in items_axes]
        self.text_axes = [self.button_X]+[self.button_Y] + [self.item_axis_label] + self.text_axes
        self.item_legend_label = Label(value="Legend position:",layout=Layout(width='16%'))
        self.item_blank_label = Label(value="",layout=Layout(width='20%'))
        
        group1 = VBox([self.buttons_view[0]],layout=Layout(width='33%'))
        group2 = VBox([self.buttons_view[1]],layout=Layout(width='33%'))
        group3 = VBox([self.buttons_view[2],self.buttons_match],layout=Layout(width='33%'))
        
        # ASSEMBLE
        display(HBox(self.buttons_load))
        self.p = plotting.PlotData()
        display(HBox(self.text_axes))
        display(HBox([self.item_blank_label,self.item_legend_label]+self.buttons_legend+[self.buttons_legend_toggle]))
        display(HBox([group1,group2,group3]))
        display(HBox(self.buttons_save))
        
        
        # Make interactive
        self.text_axes[3].observe(self.xmin, names='value')
        self.text_axes[4].observe(self.xmax, names='value')
        self.text_axes[5].observe(self.ymin, names='value')
        self.text_axes[6].observe(self.ymax, names='value')
        
        self.button_X.on_click(self.auto_x)
        self.button_Y.on_click(self.auto_y)
        
        self.buttons_legend[0].on_click(self.legend_left)
        self.buttons_legend[1].on_click(self.legend_onoff)
        self.buttons_legend[2].on_click(self.legend_right)
        self.buttons_legend_toggle.observe(self.legend_toggle)
        
        self.buttons_load[0].on_click(self.load_data)
        self.buttons_load[1].on_click(self.undo)
        
        self.buttons_view[0].on_click(self.time)
        self.buttons_view[1].on_click(self.fft)
        self.buttons_view[2].on_click(self.tf)
        
        self.buttons_match.on_click(self.match)
        
        self.buttons_save[0].on_click(self.save_data)
        self.buttons_save[1].on_click(self.save_fig)
        
        self.refresh_buttons()
        
        # Put output text at bottom
        display(self.out)
        
    
    def xmin(self,v):
        xmin = self.text_axes[3].value
        xlim = self.p.ax.get_xlim()
        self.p.ax.set_xlim([xmin,xlim[1]])
        
    def xmax(self,v):
        xmax = self.text_axes[4].value
        xlim = self.p.ax.get_xlim()
        self.p.ax.set_xlim([xlim[0],xmax])
        
    def ymin(self,v):
        ymin = self.text_axes[5].value
        ylim = self.p.ax.get_ylim()
        self.p.ax.set_ylim([ymin,ylim[1]])
        
    def ymax(self,v):
        ymax = self.text_axes[6].value
        ylim = self.p.ax.get_ylim()
        self.p.ax.set_ylim([ylim[0],ymax])
        
    def auto_x(self,b):
        self.p.auto_x()
        
    def auto_y(self,b):
        self.p.auto_y()
        
    def legend_left(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.legend_loc = 'lower left'
            self.p.update_legend(self.legend_loc)
        
    def legend_right(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.legend_loc = 'lower right'
            self.p.update_legend(self.legend_loc)
            
    def legend_onoff(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            visibility = self.p.ax.get_legend().get_visible()
            self.p.ax.get_legend().set_visible(not visibility)
                
    def legend_toggle(self,b):
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out_top:
            self.p.ax.get_legend().set_visible(True)
            self.p.legend.set_draggable(self.buttons_legend_toggle.value)
        
        
    def load_data(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        self.out_top.clear_output(wait=False)
        with self.out:
            d = file.load_data()
            if d is not None:
                self.dataset.add_to_dataset(d.time_data_list)
                self.dataset.add_to_dataset(d.freq_data_list)
                self.dataset.add_to_dataset(d.tf_data_list)
                self.dataset.add_to_dataset(d.cross_spec_data_list)
                self.dataset.add_to_dataset(d.sono_data_list)
            else:
                print('No data loaded')
            
            if len(self.dataset.time_data_list) is not 0:
                self.p.update(self.dataset.time_data_list,sets='all',channels='all')
            elif len(self.dataset.freq_data_list) is not 0:
                self.p.update(self.dataset.freq_data_list,sets='all',channels='all')
            elif len(self.dataset.tf_data_list) is not 0:
                self.p.update(self.dataset.tf_data_list,sets='all',channels='all')
            else:
                print('No data to view')
           
            self.refresh_buttons()
        
            self.p.auto_x()
            self.p.auto_y()
            
            xlim = self.p.ax.get_xlim()
            ylim = self.p.ax.get_ylim()
            self.text_axes[3].value = xlim[0]
            self.text_axes[4].value = xlim[1]
            self.text_axes[5].value = ylim[0]
            self.text_axes[6].value = ylim[1]
        
        
    def undo(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.dataset = datastructure.DataSet()
        self.out.clear_output(wait=False)
        with self.out:
            N = len(self.dataset.time_data_list)
            self.p.update(self.dataset.time_data_list,sets=[N-1],channels='all')
        

        
    def time(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        with self.out:
            N = len(self.dataset.time_data_list)
            if N is not 0:
                self.p.update(self.dataset.time_data_list)
                if self.current_view is not 'Time':
                    self.current_view = 'Time'
                    self.p.auto_x()
                    self.p.auto_y()

                xlim = self.p.ax.get_xlim()
                ylim = self.p.ax.get_ylim()
                self.text_axes[3].value = xlim[0]
                self.text_axes[4].value = xlim[1]
                self.text_axes[5].value = ylim[0]
                self.text_axes[6].value = ylim[1]
            else:
                print('no time data to display')
            
    def fft(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        with self.out:
            N = len(self.dataset.freq_data_list)
            if N is not 0:
                self.p.update(self.dataset.freq_data_list)
                if self.current_view is not 'FFT':
                    self.current_view = 'FFT'
                    self.p.auto_x()
                    self.p.auto_y() 
                    
                xlim = self.p.ax.get_xlim()
                ylim = self.p.ax.get_ylim()
                self.text_axes[3].value = xlim[0]
                self.text_axes[4].value = xlim[1]
                self.text_axes[5].value = ylim[0]
                self.text_axes[6].value = ylim[1]
            else:
                print('no FFT data to display')
    
    def tf(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        with self.out:
            N = len(self.dataset.tf_data_list)
            if N is not 0:
                self.p.update(self.dataset.tf_data_list)
                if self.current_view is not 'TF':
                    self.current_view = 'TF'
                    self.p.auto_x()
                    self.p.auto_y()
                xlim = self.p.ax.get_xlim()
                ylim = self.p.ax.get_ylim()
                self.text_axes[3].value = xlim[0]
                self.text_axes[4].value = xlim[1]
                self.text_axes[5].value = ylim[0]
                self.text_axes[6].value = ylim[1]
            else:
                print('no TF data to display')
                
    def match(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        with self.out:
            if self.current_view is 'TF':
                freq_range = self.p.ax.get_xlim()
                current_calibration_factors = self.dataset.tf_data_list.get_calibration_factors()
                reference = current_calibration_factors[0][0]
                factors = analysis.best_match(self.dataset.tf_data_list,freq_range=freq_range,set_ref=0,ch_ref=0)
                factors = [reference*x for x in factors]
                self.dataset.tf_data_list.set_calibration_factors_all(factors)
                self.p.update(self.dataset.tf_data_list)
                print('scale factors:')
                print(factors)
                #self.p.auto_y()
            else:
                print('First press <View TF>')
            
    
    
    def save_data(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        with self.out:
            print('Saving dataset:')
            print(self.dataset)
            self.dataset.save_data()
            
    
    def save_fig(self,b):
        # the 'out' construction is to refresh the text output at each update 
        # to stop text building up in the widget display
        self.out.clear_output(wait=False)
        with self.out:
            file.save_fig(self.p,figsize=(9,5))
    
    def refresh_buttons(self):
        if len(self.dataset.time_data_list) is 0:
            self.buttons_view[0].button_style =''
        else:
            self.buttons_view[0].button_style ='info'
            
        if len(self.dataset.freq_data_list) is 0:
            self.buttons_view[1].button_style =''
        else:
            self.buttons_view[1].button_style ='info'
            
        if len(self.dataset.tf_data_list) is 0:
            self.buttons_view[2].button_style =''
        else:
            self.buttons_view[2].button_style ='info'
                


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 19 17:29:30 2019

@author: tb267
"""


from . import datastructure
from . import options

import numpy as np
from scipy import signal
from scipy import optimize
import copy

MESSAGE = ''

#%% single peak fit

def f_3dB(f,G0):
    
    ff = np.linspace(f[0],f[-1],np.max([len(f),1000]))
    GG = np.interp(ff,f,np.squeeze(G0))
    
    f=ff
    G0=GG
    
    
    fn0i = np.argmax(np.abs(G0))
    fn0 = f[fn0i]
    
    halfpower = np.max(np.abs(G0)) / np.sqrt(2)
    
    xi = np.where(np.squeeze(np.abs(G0)) > halfpower)[0]

    
    f1 = f[xi[0]]
    f2 = f[xi[-1]]

    
    df = f2-f1
    zn0 = df/2/fn0
    
    return fn0,zn0



def modal_fit_single_channel(tf_data,freq_range=None,channel=0,measurement_type='acc'):
    '''
    Fit modal parameters for a single mode to data within specified freq_range
    '''
    
    if freq_range == None:
        freq_range = tf_data.freq_axis[[0,-1]]
    
    
    f = tf_data.freq_axis
    selected_range = np.where((f > freq_range[0]) & (f < freq_range[1]))
    
    f = f[selected_range]
    
    #NEED GOOD INITIAL GUESS!
    G0 = tf_data.tf_data[selected_range,channel]
    
    fn0,zn0 = f_3dB(f,G0)
    
    
    
    an0 = np.max(np.abs(G0))*(2*np.pi*fn0)**2 * 2*zn0
    pn0 = 0
    Rk0 = np.max(np.abs(G0))/1e3
    Rm0 = np.max(np.abs(G0))*((2*np.pi*fn0)**2)/1e3
    
    
    x0 = np.array([fn0,zn0,an0,pn0,Rk0,Rm0])
    
    bounds = ([freq_range[0],0,-np.inf,-np.pi/2,0,0],[freq_range[1],1,np.inf,np.pi/2,np.inf,np.inf])
#    bounds = ((-np.inf,np.inf),(freq_range[0],freq_range[1]),(0,np.inf),(-np.pi/2,np.pi/2),(-np.inf,np.inf),(-np.inf,np.inf))
    
    r = optimize.least_squares(f_residual,x0, bounds=bounds, max_nfev=1000, args=(f,G0,measurement_type))
#    r = optimize.minimize(f_residual,x0, bounds=bounds, method='SLSQP', args=(f,G0,measurement_type))
    print('')
    print('fn={:.4g} (Hz), zn={:.4g}, an={:.4g}, phase={:.4g} (deg)'.format(r.x[0],r.x[1],r.x[2],r.x[3]*180/np.pi))
    print('')
    if (np.abs(r.x[3])*180/np.pi > 60):
        print('Phase is significant, check ''measurement_type'' setting is correct')
    return r
    
    

def f_TF(x,f,measurement_type):

    fn = x[0]
    zn = x[1]
    an = x[2]
    pn = x[3]
#    R  = x[4] + 1j*x[5]
    R1 = x[4]
    R2 = x[5]
    
    wn = 2*np.pi*fn
    w = 2*np.pi*f
    
    
    if measurement_type == 'acc':
        p = 2
    elif measurement_type == 'vel':
        p = 1
    elif measurement_type == 'dsp':
        p = 0
    
    G = an*np.exp(1j*pn)/(wn**2 + 2j*wn*zn*w - w**2) + R1 - R2/(w**2)
    G = G*((1j*w)**p)
    
    return G



def f_residual(x,f,G0,measurement_type):
    
    G0 = np.squeeze(G0)
    G = f_TF(x,f,measurement_type)
        
    e = (G-G0)
    #e = np.abs(e)
    e = np.concatenate((np.real(e),np.imag(e)))
    
    
    return e



#%%
def f_TF_all_channels(x,f,measurement_type):

    N_tfs = np.int((len(x)-2)/4)
    
    fn = x[0]
    zn = x[1]
    an = x[2:2+N_tfs].reshape(1,N_tfs)
    pn = x[2+N_tfs:2+2*N_tfs].reshape(1,N_tfs)
#    R  = x[4] + 1j*x[5]
    R1 = x[2+2*N_tfs:2+3*N_tfs].reshape(1,N_tfs)
    R2 = x[2+3*N_tfs:2+4*N_tfs].reshape(1,N_tfs)
    
    wn = 2*np.pi*fn
    w = 2*np.pi*f
    w = w.reshape(len(w),1)
    
    
    if measurement_type == 'acc':
        p = 2
    elif measurement_type == 'vel':
        p = 1
    elif measurement_type == 'dsp':
        p = 0

    G = an*np.exp(1j*pn)/(wn**2 + 2j*wn*zn*w - w**2) + R1 - R2/(w**2)
    G = G*((1j*w)**p)
    
    
    
    return G


def f_residual_all_channels(x,f,G0,measurement_type):
    
    G = f_TF_all_channels(x,f,measurement_type)
        
    e = (G-G0)
    #e = np.abs(e)
    e = np.concatenate((np.real(e),np.imag(e)))
    e = e.reshape(np.size(e))
    
    return e

#%% MULTI-CHANNEL MODAL FIT
def modal_fit_all_channels(tf_data_list,freq_range=None,measurement_type='acc'):
    '''
    Fit modal parameters for a single mode to data within specified freq_range.
    
    Assumes all tf_data in tf_data_list have same frequency axes
    '''
    global MESSAGE
    
    if measurement_type == 'acc':
        p = 2
    elif measurement_type == 'vel':
        p = 1
    elif measurement_type == 'dsp':
        p = 0
    
    # Find out how many TFs in dataset
    N_tfs = 0
    for tf_data in tf_data_list:
        if tf_data.flag_modal_TF == False:
            N_tfs += len(tf_data.tf_data[0,:])
    
    if freq_range == None:
        freq_range = tf_data.freq_axis[[0,-1]]
    
    # get selected frequency axis
    f = tf_data_list[0].freq_axis
    selected_range = np.where((f > freq_range[0]) & (f < freq_range[1]))
    
    f = f[selected_range]
    
    # compile transfer functions into single array, and get initial guesses
    G0 = np.zeros((len(f),N_tfs),dtype=complex)
    fn0 = np.zeros(N_tfs)
    zn0 = np.zeros(N_tfs)
    counter = -1
    for tf_data in tf_data_list:
        if tf_data.flag_modal_TF == False:
            for n_chan in range(len(tf_data.tf_data[0,:])):
                counter += 1
                G0[:,counter] = tf_data.tf_data[selected_range,n_chan] * tf_data.channel_cal_factors[n_chan]
                fn0[counter],zn0[counter] = f_3dB(f,G0[:,counter])
            

    # initial global guess for fn0,zn0 discarding any outliers
    fn0 = np.median(fn0)
    zn0 = np.median(zn0)
    fn0i = np.argmin(np.abs(f - fn0))
    
    
    # initial guesses for each channel
    an0 = np.zeros(N_tfs)
    pn0 = np.zeros(N_tfs)
    Rk0 = np.zeros(N_tfs)
    Rm0 = np.zeros(N_tfs)
    id_link = []
    counter = -1
    for tf_data in tf_data_list:
        if tf_data.flag_modal_TF == False:
            id_link += [tf_data.id_link]
            for n_chan in range(len(tf_data.tf_data[0,:])):
                counter += 1
                an0[counter] = np.max(np.abs(G0[:,counter]))*(2*np.pi*fn0)**(2-p) * 2*zn0
    #            an0[counter] = an0[counter] * np.sign(np.real(G0[fn0i,counter] / ((2j*np.pi*fn0)**p)))
                an0[counter] = an0[counter] * np.sign(-np.imag(G0[fn0i,counter] / ((1j)**p)))
                
                pn0[counter] = 0
                Rk0[counter] = np.max(np.abs(G0[:,counter]))/1e6
                Rm0[counter] = np.max(np.abs(G0[:,counter]))*((2*np.pi*fn0)**2)/1e6
    
    x0 = np.concatenate(([fn0],[zn0],an0,pn0,Rk0,Rm0))
    
    
    # bounds
    B_fn = freq_range
    B_zn = [0,1]
    B_an = np.zeros((N_tfs,2))
    B_an[:,0] = -np.inf
    B_an[:,1] =  np.inf
    B_pn = np.zeros((N_tfs,2))
    B_pn[:,0] = -np.pi/2
    B_pn[:,1] =  np.pi/2
    B_rk = np.zeros((N_tfs,2))
    B_rk[:,0] = 0
    B_rk[:,1] = np.inf
    B_rm = np.copy(B_rk)
    
    lower_bounds = np.concatenate(([B_fn[0]],[B_zn[0]],B_an[:,0],B_pn[:,0],B_rk[:,0],B_rm[:,0]))
    upper_bounds = np.concatenate(([B_fn[1]],[B_zn[1]],B_an[:,1],B_pn[:,1],B_rk[:,1],B_rm[:,1]))
    
    bounds = (lower_bounds,upper_bounds)
    
    r = optimize.least_squares(f_residual_all_channels,x0, bounds=bounds, max_nfev=1000, args=(f,G0,measurement_type))
    
    settings = tf_data_list[0].settings
    test_name = tf_data_list[0].test_name
    m = datastructure.ModalData(r.x, settings=settings, id_link=id_link, test_name=test_name)

    fn,zn,an,pn,rk,rm = unpack(r.x)

#    with np.printoptions(precision=3, suppress=True):
    MESSAGE = 'fn={:.4g} (Hz), zn={:.4g}\n\n'.format(fn,zn)
    MESSAGE += 'an={}\n\n'.format(an)
    MESSAGE += 'pn={} deg\n\n'.format(pn*180/np.pi)
    print(MESSAGE)
    if np.any(np.abs(pn)*180/np.pi > 60):
        MESSAGE += '\nPhase is significant, check ''TF type'' setting is correct.\n'
        print(MESSAGE)
        
    # Check quality of fit
    e = f_residual_all_channels(r.x,f,G0,measurement_type)
    G0rms = np.mean(np.abs(G0)**2)
    erms = np.mean(np.abs(e)**2)
    e_rel = erms/G0rms
    if e_rel > 0.1:
        MESSAGE += '\nPoor quality fit: try adjusting frequency range.\n'
        print(MESSAGE)

    return m
    
#%%% Reconstruction
def unpack(x):
    # unpacks modal parameters into set of variables
    N_tfs = np.int((len(x)-2)/4)
    
    fn = x[0]
    zn = x[1]
    an = x[2:2+N_tfs]
    pn = x[2+N_tfs:2+2*N_tfs]
    rk = x[2+2*N_tfs:2+3*N_tfs]
    rm = x[2+3*N_tfs:2+4*N_tfs]
    
    return fn,zn,an,pn,rk,rm

def unpack_matrix(X):
    # unpacks modal parameters into set of variables
    N_tfs = np.int((len(X[0,:])-2)/4)

    fn = X[:,0]
    zn = X[:,1]
    an = X[:,2:2+N_tfs]
    pn = X[:,2+N_tfs:2+2*N_tfs]
    rk = X[:,2+2*N_tfs:2+3*N_tfs]
    rm = X[:,2+3*N_tfs:2+4*N_tfs]

    return fn,zn,an,pn,rk,rm
    
def pack(fn,zn,an,pn,rk,rm):
    # packs modal parameters into single variable for optimisation
    x = np.concatenate(([fn],[zn],an,pn,rk,rm))
    return x
    

def reconstruct_transfer_function(modal_data,f,measurement_type='acc'):
    '''
    Reconstructs transfer functions from modal_data and returns TfData object
    '''
    G = 0
    for n_row in range(len(modal_data.M[:,0])):
        xn = modal_data.M[n_row,:]
        G += f_TF_all_channels(xn,f,measurement_type=measurement_type)
    
    settings = modal_data.settings
    settings.channels = modal_data.channels
    tf_data = datastructure.TfData(f,G,None,settings,units=modal_data.units,channel_cal_factors=None,id_link=modal_data.id_link,test_name=modal_data.test_name)
    tf_data.flag_modal_TF = True
    return tf_data

def reconstruct_transfer_function_global(modal_data,f,measurement_type='acc'):
    '''
    Reconstructs transfer functions from modal_data and returns TfData object
    '''
    G = 0
    N_tfs = np.int((len(modal_data.M[0,:])-2)/4)
    for n_row in range(len(modal_data.M[:,0])):
        xn = modal_data.M[n_row,:]
        xn[2+2*N_tfs:] = 0 #don't want local residual fits for global fits - i.e. rk and rm
        G += f_TF_all_channels(xn,f,measurement_type=measurement_type)
    
    settings = modal_data.settings
    settings.channels = modal_data.channels
    tf_data = datastructure.TfData(f,G,None,settings,units=modal_data.units,channel_cal_factors=None,id_link=modal_data.id_link,test_name=modal_data.test_name)
    tf_data.flag_modal_TF = True
    return tf_data


# -*- coding: utf-8 -*-
"""
Created on Tue Aug 28 19:04:14 2018

@author: tb267
"""



from . import options
from . import datastructure

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
matplotlib.rcParams.update({'font.size': 12,'font.family':'serif'})
from matplotlib.ticker import AutoLocator

LINE_ALPHA = 0.9

class PlotSonoData():
    def __init__(self,figsize=(9,5),canvas=None,fig=None):
        if canvas==None:
            self.fig, self.ax = plt.subplots(1,1,figsize=figsize,dpi=100)
        else:
            self.fig = fig
            self.canvas = canvas
            self.ax = self.canvas.figure.subplots()
        
        self.ax.grid(False,alpha=0.2)
        self.fig.canvas.draw()
        
    def update(self,sono_data_list,n_set=0,n_chan=0):
        
        f = sono_data_list[n_set].freq_axis
        t = sono_data_list[n_set].time_axis
        S = sono_data_list[n_set].sono_data[:,:,n_chan]
        self.ax.pcolor(t,f,20*np.log10(np.abs(S)))
        self.ax.set_xlabel('Time (s)')
        self.ax.set_ylabel('Frequency (Hz)')
    

class PlotData():
    def __init__(self,sets='all',channels='all',figsize=(9,5),canvas=None,fig=None):
        if canvas==None:
            self.fig, self.ax = plt.subplots(1,1,figsize=figsize,dpi=100)
        else:
            self.fig = fig
            self.canvas = canvas
            self.ax = self.canvas.figure.subplots()

        #ax2 is for coherence
        self.ax2 = self.ax.twinx()
        self.ax2.set_visible(False)
        self.ax.set_zorder(self.ax2.get_zorder()+1)
        self.ax.patch.set_visible(False)
        
#        #ax3 is for sonograms: 
#        self.ax2 = self.ax.twinx()
#        self.ax2.set_visible(False)
#        self.ax.set_zorder(self.ax2.get_zorder()+1)
#        self.ax.patch.set_visible(False)

        self.ax.lines=[]
        self.ax2.lines=[]
        self.line_listbyset = []
        self.line2_listbyset = []
        self.pcolor_sono = None
        self.visibility = True
        
        self.ax.grid(True,alpha=0.2)
        self.fig.canvas.mpl_connect('pick_event', self.channel_select)
        self.fig.canvas.draw()
        
    def update(self,data_list,sets='all',channels='all',xlinlog='linear',show_coherence=True,plot_type=None,coherence_plot_type='linear',freq_range=None, auto_xy=''):
        global LINE_ALPHA
        
        # when switching back from sonogram, remove all pcolormesh parts of plot
        for ch in self.ax.get_children():
                if ch.__class__.__name__ == 'QuadMesh':
                    ch.remove()
                    self.visibility = True # turn legend back on
        
        
        self.ax.grid(True,alpha=0.2)
        self.data_list = data_list
        self.plot_type = plot_type
        self.freq_range = freq_range
        self.xlinlog = xlinlog
        
        if data_list.__class__.__name__ == 'TimeDataList':
            self.ax2.set_visible(False)
            self.ax.set_xlabel('Time (s)')
            self.ax.set_ylabel('Amplitude')
        elif data_list.__class__.__name__ == 'FreqDataList':
            self.ax2.set_visible(False)
            if (plot_type == 'Amplitude (dB)') or (plot_type == None):
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Amplitude (dB)')
            elif plot_type == 'Amplitude (linear)':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Amplitude')
            elif plot_type == 'Real Part':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Real Part')
            elif plot_type == 'Imag Part':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Imag Part')
            elif plot_type == 'Nyquist':
                self.ax.set_xlabel('Real Part')
                self.ax.set_ylabel('Imag Part')
            elif plot_type == 'Phase':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Phase (deg)')
        
                
            self.ax.set_xscale(xlinlog)
            if 'log' in xlinlog:
                self.ax.grid(b=True, which='minor',axis='x',alpha=0.2)
            else:
                self.ax.grid(b=False,which='minor',axis='x',alpha=0.2)
            
        elif data_list.__class__.__name__ == 'TfDataList':
            if (plot_type == 'Amplitude (dB)') or (plot_type == None):
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Amplitude (dB)')
            elif plot_type == 'Amplitude (linear)':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Amplitude')
            elif plot_type == 'Real Part':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Real Part')
            elif plot_type == 'Imag Part':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Imag Part')
            elif plot_type == 'Nyquist':
                self.ax.set_xlabel('Real Part')
                self.ax.set_ylabel('Imag Part')
            elif plot_type == 'Phase':
                self.ax.set_xlabel('Frequency (Hz)')
                self.ax.set_ylabel('Phase (deg)')
            
            
            self.ax.set_xscale(xlinlog)
            if 'log' in xlinlog:
                self.ax.grid(b=True, which='minor',axis='x',alpha=0.2)
            else:
                self.ax.grid(b=False,which='minor',axis='x',alpha=0.2)
            # setup twin axis
            
            if show_coherence == True:
                self.ax2.set_ylabel('Coherence')
                self.ax2.set_visible(True)
            else:
                self.ax2.set_ylabel('')
    
        # sonogram plot completely different, use separate function
        
        N_sets = len(data_list)

        # pre-count how many channels total
        ch_counter = 0
        for n_set in range(N_sets):
            if data_list.__class__.__name__ == 'TimeDataList':
                ch_counter += len(data_list[n_set].time_data[0,:])
            elif data_list.__class__.__name__ == 'FreqDataList':
                ch_counter += len(data_list[n_set].freq_data[0,:])
            elif data_list.__class__.__name__ == 'TfDataList':
                ch_counter += len(data_list[n_set].tf_data[0,:])

        self.ch_total = ch_counter
        if self.ch_total <= 10:
            LINE_ALPHA = 0.9
        else:
            LINE_ALPHA = 1-1/self.ch_total # make deselected lines fainter if more channels
        
        self.ax.lines=[]
        self.ax2.lines=[]
        self.line_listbyset = []
        self.line2_listbyset = []
        
        if sets == 'all':
            sets = range(N_sets)
        count = -1
        for n_set in range(len(data_list)):
            self.line_listbyset.append([])
            self.line2_listbyset.append([])
            
            if data_list.__class__.__name__ == 'TfDataList':
                N_chans = len(data_list[n_set].tf_data[0,:])
            elif data_list.__class__.__name__ == 'FreqDataList':
                N_chans = len(data_list[n_set].freq_data[0,:])
            elif data_list.__class__.__name__ == 'TimeDataList':
                N_chans = len(data_list[n_set].time_data[0,:])
                
            if channels == 'all':
                channels = range(N_chans)

            

            for n_chan in range(N_chans):
                count += 1
                if (n_set not in sets) or (n_chan not in channels):
                    alpha = 1-LINE_ALPHA
                else:
                    alpha = LINE_ALPHA
                
                if data_list.__class__.__name__ == 'TimeDataList':
                    x = data_list[n_set].time_axis
                    y = data_list[n_set].time_data[:,n_chan] * data_list[n_set].channel_cal_factors[n_chan]
        
                elif data_list.__class__.__name__ == 'FreqDataList':
                    x = data_list[n_set].freq_axis
                    ylin = data_list[n_set].freq_data[:,n_chan] * data_list[n_set].channel_cal_factors[n_chan]
                    
                    
                    if (plot_type == 'Amplitude (dB)') or (plot_type == None):
                        # handle log(0) manually to avoid warnings
                        y = np.zeros(np.shape(ylin))
                        izero = ylin==0
                        y[~izero] = 20*np.log10(np.abs(ylin[~izero]))
                        y[izero] = -np.inf
                    elif plot_type == 'Amplitude (linear)':
                        y = np.abs(ylin)
                    elif plot_type == 'Real Part':
                        y = np.real(ylin)
                    elif plot_type == 'Imag Part':
                        y = np.imag(ylin)
                    elif plot_type == 'Nyquist':
                        if freq_range == None:
                            freq_range = [-1,np.inf]
                        selected_data = np.where((x>freq_range[0]) * (x<freq_range[1]))[0]
                        x = np.real(ylin[selected_data])
                        y = np.imag(ylin[selected_data])
                        
                    elif plot_type == 'Phase':
                        y = np.angle(ylin,deg=True)
                        
                    
                elif data_list.__class__.__name__ == 'TfDataList':
                    x = data_list[n_set].freq_axis
                    ylin = data_list[n_set].tf_data[:,n_chan] * data_list[n_set].channel_cal_factors[n_chan]
                    if (plot_type == 'Amplitude (dB)') or (plot_type == None):
                        # handle log(0) manually to avoid warnings
                        y = np.zeros(np.shape(ylin))
                        izero = ylin==0
                        y[~izero] = 20*np.log10(np.abs(ylin[~izero]))
                        y[izero] = -np.inf
                        
                    elif plot_type == 'Amplitude (linear)':
                        y = np.abs(ylin)
                    elif plot_type == 'Real Part':
                        y = np.real(ylin)
                    elif plot_type == 'Imag Part':
                        y = np.imag(ylin)
                    elif plot_type == 'Nyquist':
                        if freq_range == None:
                            freq_range = [-1,np.inf]
                        selected_data = np.where((x>freq_range[0]) * (x<freq_range[1]))[0]
                        x = np.real(ylin[selected_data])
                        y = np.imag(ylin[selected_data])
                    elif plot_type == 'Phase':
                        y = np.angle(ylin,deg=True)
                    if data_list[n_set].tf_coherence is not None:
                        # handle log(0) manually to avoid warnings
                        yclin = data_list[n_set].tf_coherence[:,n_chan]
                        if coherence_plot_type == 'linear':
                            yc = yclin
                        elif coherence_plot_type == 'log':
                            yc = np.zeros(np.shape(yclin))
                            izero = yclin==0
                            yc[~izero] = 20*np.log10(np.abs(yclin[~izero]))
                            yc[izero] = -np.inf
                            yc = 20*np.log10(np.abs(yclin))
                    else:
                        yc = np.ones(np.shape(data_list[n_set].tf_data))
                        
                        
                color = options.set_plot_colours(ch_counter)[count,:]/255
                
                if type(data_list[n_set].test_name) is str:
                    test_name = data_list[n_set].test_name
                else:
                    test_name = 'set '
                
                if self.ch_total < 10:
                    label = '{}_{}, ch_{}'.format(test_name,n_set,n_chan)
                else:
                    label = 'set{},ch{}'.format(n_set,n_chan)
                    
                if data_list.__class__.__name__ == 'TfDataList':
                    if data_list[n_set].flag_modal_TF == True:
                        label = 'fit{}'.format(n_chan)
                        
                self.line_listbyset[n_set] += self.ax.plot(x,y,'-',linewidth=1,color = color,label=label,alpha=alpha)
                
                if data_list.__class__.__name__ == 'TfDataList':
                    if show_coherence == True:
                        self.line2_listbyset[n_set] += self.ax2.plot(x,yc,':',linewidth=1,color = color,label=label+' (coherence)',alpha=alpha)
                        self.ax2.set_visible(True)
                    else:
                        self.ax2.set_visible(False)
        
        
        self.update_legend()
        if 'x' in auto_xy:
            self.auto_x()
        if 'f' in auto_xy: # use freq_range for x axis
            if ('log' in xlinlog) and freq_range[0]==0:
                self.lines = self.ax.get_lines()
                xminlog = np.inf
                for line in self.lines:
                    if line.get_alpha() > 0.5:
                        data = line.get_data()
                        xxlog = data[0][1] # first nonzero freq axis
                        xminlog = min([xminlog,xxlog])
                        freq_range[0] = xminlog
                self.ax.set_xlim(freq_range)
            else:
                self.ax.set_xlim(freq_range)
            self.fig.canvas.draw()
            self.auto_y()
        if 'y' in auto_xy:
            self.auto_y()
        if 'c' in auto_xy: # auto coherence y
            self.ax2.set_ylim([0,1])
            
#        self.ax.xaxis.set_major_locator(AutoLocator())
#        self.ax.yaxis.set_major_locator(AutoLocator())
        self.fig.canvas.draw()
        
        
    def update_legend(self,loc='lower right',draggable=False):
        
        if len(self.data_list) != 0:
            if self.ax.get_legend() is None:
                self.visibility = True
                
            if self.ch_total >= 10:
                # make legend more compact
                col_sizes = np.arange(10,7,-1)
                rem = np.remainder(self.ch_total,col_sizes)
                if any(rem==0):
                    remi = np.where(rem==0)[0]
                    remi = remi[0]
                    ncol = np.int(self.ch_total / col_sizes[remi])
                else:
                    remi = np.argmax(rem)
                    ncol = np.int(np.ceil(self.ch_total/ col_sizes[remi]))
            else:
                ncol = 1
                
            self.legend = self.ax.legend(loc=loc, ncol=ncol)
#            self.ax.legend()
            self.legend.set_draggable(draggable,use_blit=True)#(True),update='bbox',use_blit=True
            self.lines = self.ax.get_lines()
            self.lines2 = self.ax2.get_lines()
            self.lined = dict()
            self.lined2 = dict()

            # make dictionary of legend lines for selection    
            for legline, origline in zip(self.legend.get_lines(), self.lines):
                legline.set_picker(10)  # argument tolerance
                self.lined[legline] = origline
                legline.set_alpha(origline.get_alpha())

            # make dictionary of coherence lines to select with legend - only for TF Data
            if len(self.ax2.lines) > 0:
                for legline, origline2 in zip(self.legend.get_lines(), self.lines2):
                    self.lined2[legline] = origline2 
                    origline2.set_alpha(legline.get_alpha())
                    
            self.ax.get_legend().set_visible(self.visibility)
        else:
            self.legend = self.ax.get_legend()
            if self.legend is not None:
                self.legend.remove()
            self.fig.canvas.draw()
    
    
    def auto_x(self):
        if self.data_list.__class__.__name__ == 'SonoDataList':
            xlim = self.data_list[self.n_set].time_axis[[0,-1]]
            self.ax.set_xlim(xlim)
        else:
            self.lines = self.ax.get_lines()
            xmin = np.inf
            xminlog = np.inf
            xmax = -np.inf
            for line in self.lines:
                if line.get_alpha() > 0.5:
                    data = line.get_data()
                    xx = min(data[0])
                    xmin = min([xx,xmin])
                    
                    xxlog = data[0][1] # first nonzero freq axis
                    xminlog = min([xminlog,xxlog])
                    
                    xx = max(data[0])
                    xmax = max([xx,xmax])
            try:
                if 'log' in self.xlinlog:
                    self.ax.set_xlim([xminlog,xmax])
                else:
                    self.ax.set_xlim([xmin,xmax])
            except:
                pass
        self.fig.canvas.draw()
        
        
    def auto_y(self):
        if self.data_list.__class__.__name__ == 'SonoDataList':
            ylim = self.data_list[self.n_set].freq_axis[[0,-1]]
            self.ax.set_ylim(ylim)
        else:
            self.lines = self.ax.get_lines()
            if (self.data_list.__class__.__name__ == 'TfData') and (self.plot_type != 'Nyquist'):
                xview = self.freq_range
            else:
                xview = self.ax.get_xlim()
            ymin = np.inf
            ymax = -np.inf
            c=-1
            for line in self.lines:
                c+=1
                data = line.get_data() 
                x = data[0]
                selection = (xview[0] < x) & (x < xview[1])
                if line.get_alpha() > 0.5:
                    yy = min(data[1][selection])
                    ymin = min([yy,ymin])
                    yy = max(data[1][selection])
                    ymax = max([yy,ymax])
            try:
                self.ax.set_ylim([ymin,ymax])
            except:
                pass
        self.fig.canvas.draw()
        
    def channel_select(self,event):
        selected_line = event.artist
        
        a = selected_line.get_alpha()
        # This function is called when legend dragged, and a is None
        # So only want to change line opacity when line actually selected.
        if a is not None:
            # change opacity of both legend line and actual line
            a = 1-a
            origline = self.lined[selected_line]
            

            selected_line.set_alpha(a)
            origline.set_alpha(a)
            
            # change z order to bring selected line to foreground
            for line in self.ax.lines:
                line.set_zorder(0)
            if a > 0.5:
                origline.set_zorder(1)
            else:
                origline.set_zorder(0)
                
            # also select matching coherence lines
            if len(self.ax2.lines) > 0:
                origline2 = self.lined2[selected_line]
                origline2.set_alpha(a)
                for line2 in self.ax2.lines:
                    line2.set_zorder(0)
                if a > 0.5:
                    origline2.set_zorder(1)
                else:
                    origline2.set_zorder(0)
            self.fig.canvas.draw()
    
    def get_selected_channels(self):
        # find the sets and channels higlighted in figure
        
        n_sets = len(self.line_listbyset)
        selected_data = []
        
        for ns in range(n_sets):
            selected_data += [[]]
            for line in self.line_listbyset[ns]:
                selected_data[ns] += [line.get_alpha() > 0.5] # skip coherence lines
            
        return selected_data
    
    def set_selected_channels(self,s):
        global LINE_ALPHA
        # relies on all sets of data with same number of channels. Need to make more general.
        for n_set in range(len(s)):
            for n_chan in range(len(s[n_set])):
                if s[n_set][n_chan] == True:
                    self.line_listbyset[n_set][n_chan].set_alpha(LINE_ALPHA)
                else:
                    self.line_listbyset[n_set][n_chan].set_alpha(1-LINE_ALPHA)
        # make legend line alphas match actual lines, but keep visibility as before
        self.update_legend()
        self.fig.canvas.draw()
        
    def update_sonogram(self,sono_data_list,n_set,n_chan,db_range=60,auto_xy='xy'):
        self.data_list = sono_data_list # makes auto_x/y work!
        
        for ch in self.ax.get_children():
                if ch.__class__.__name__ == 'QuadMesh':
                    ch.remove()
        
        self.n_set = n_set
        self.n_chan = n_chan
        self.ax2.set_visible(False)
        self.ax.set_xlabel('Time (s)')
        self.ax.set_ylabel('Frequency (Hz)')
        self.ax.grid(False,alpha=0.2)
        if self.ax.get_legend() is not None:
            self.ax.get_legend().set_visible(False)
        
        self.ax.lines=[]
        self.ax2.lines=[]
        self.line_listbyset = []
        self.line2_listbyset = []
        
        f = sono_data_list[n_set].freq_axis
        t = sono_data_list[n_set].time_axis
        S = sono_data_list[n_set].sono_data[:,:,n_chan]
        SdB = 20*np.log10(np.abs(S))
        vmax = SdB.max()
        vmin = np.max([SdB.min(),vmax-db_range])
        self.pcolor_sono = self.ax.pcolormesh(t,f,SdB,shading='gouraud',cmap='Blues',vmax=vmax,vmin=vmin,rasterized=True)
        if 'x' in auto_xy:
            self.auto_x()
        if 'y' in auto_xy:
            self.auto_y()

#        self.fig.colorbar(ax=self.ax)
        self.fig.canvas.draw()
        
        
#    def plot_modal_reconstruction(modal_data_list):
#        for data in modal_data:
#            if data.__class__.__name__ == 'TfData':
#                
#                self.ax.plot
    

#class PlotTimeData():
#    def __init__(self,time_data_list,sets='all',channels='all'):
#        self.fig, self.ax = plt.subplots(1,1,figsize=(9,5),dpi=100)
#        self.ax.set_xlabel('Time (s)')
#        self.ax.set_ylabel('Normalised Amplitude')
#        self.ax.grid(True,alpha=0.3)
#        self.update(time_data_list,sets,channels)
#        self.fig.canvas.mpl_connect('pick_event', self.channel_select)
#        self.fig.canvas.draw()
#        
#    def update(self,time_data_list,sets,channels):
#        
#        self.ax.lines=[]
#        if sets == 'all':
#            sets = range(len(time_data_list))
#        
#        count = -1
#        for n_set in range(len(time_data_list)):
#            if channels == 'all':
#                channels = range(time_data_list[n_set].settings.channels)
#        
#            for n_chan in range(time_data_list[n_set].settings.channels):
#                count += 1
#                
#                if (n_set not in sets) or (n_chan not in channels):
#                    alpha = 1-LINE_ALPHA
#                else:
#                    alpha = LINE_ALPHA
#                    
#                t = time_data_list[n_set].time_axis
#                y = time_data_list[n_set].time_data[:,n_chan]
#                color = options.set_plot_colours(len(time_data_list)*time_data_list[n_set].settings.channels)[count,:]/255
#                
#                label = 'set{},ch{}'.format(n_set,n_chan)
#                self.ax.plot(t,y,'-',linewidth=1,color = color,label=label,alpha=alpha)
#        
#        if len(time_data_list) != 0:
#            self.legend = self.ax.legend()
#            self.lines = self.ax.get_lines()
#            self.lined = dict()
#            for legline, origline in zip(self.legend.get_lines(), self.lines):
#                legline.set_picker(10)  # 5 pts tolerance
#                self.lined[legline] = origline 
#        else:
#            self.legend = self.ax.get_legend()
#            if self.legend is not None:
#                self.legend.remove()
#            self.fig.canvas.draw()
#    
#    
#    def channel_select(self,event):
#        selected_line = event.artist
#        
#        a = selected_line.get_alpha()
#        a = 1-a
#        
#        origline = self.lined[selected_line]
#        selected_line.set_alpha(a)
#        origline.set_alpha(a)
#        self.fig.canvas.draw()
#
#        
#        
#
#
#class PlotFreqData():
#    def __init__(self,freq_data_list,sets='all',channels='all'):
#        self.fig, self.ax = plt.subplots(1,1,figsize=(9,5),dpi=100)
#        self.ax.set_xlabel('Frequency (Hz)')
#        self.ax.set_ylabel('Normalised Amplitude')
#        self.ax.grid(True,alpha=0.3)
#        self.update(freq_data_list,sets,channels)
#        self.fig.canvas.mpl_connect('pick_event', self.channel_select)
#        self.fig.canvas.draw()
#        
#    def update(self,freq_data_list,sets,channels):
#        
#        self.ax.lines=[]
#        if sets == 'all':
#            sets = range(len(freq_data_list))
#        
#        count = -1
#        for n_set in range(len(freq_data_list)):
#            if channels == 'all':
#                channels = range(freq_data_list[n_set].settings.channels)
#        
#            for n_chan in range(freq_data_list[n_set].settings.channels):
#                count += 1
#                
#                if (n_set not in sets) or (n_chan not in channels):
#                    alpha = 1-LINE_ALPHA
#                else:
#                    alpha = LINE_ALPHA
#                    
#                f = freq_data_list[n_set].freq_axis
#                Y = freq_data_list[n_set].freq_data[:,n_chan]
#                color = options.set_plot_colours(len(freq_data_list)*freq_data_list[n_set].settings.channels)[count,:]/255
#                
#                label = 'set{},ch{}'.format(n_set,n_chan)
#                self.ax.plot(f,20*np.log10(np.abs(Y)),'-',linewidth=1,color = color,label=label,alpha=alpha)
#        
#        if len(freq_data_list) != 0:
#            self.legend = self.ax.legend()
#            self.lines = self.ax.get_lines()
#            self.lined = dict()
#            for legline, origline in zip(self.legend.get_lines(), self.lines):
#                legline.set_picker(10)  # 5 pts tolerance
#                self.lined[legline] = origline 
#        else:
#            self.legend = self.ax.get_legend()
#            if self.legend is not None:
#                self.legend.remove()
#            self.fig.canvas.draw()
#    
#    
#    def channel_select(self,event):
#        selected_line = event.artist
#        
#        a = selected_line.get_alpha()
#        a = 1-a
#        
#        origline = self.lined[selected_line]
#        selected_line.set_alpha(a)
#        origline.set_alpha(a)
#        self.fig.canvas.draw()
#        
#        
#
#
#
#class PlotData2(object):
#    def __init__(self,data,channels='all',plot_coherence=True):
#        '''
#        Args:
#            data: plots data which can be class of type:
#                    datastructure.DataSet
#                    datastructure.TimeData
#                    datastructure.FreqData
#                    datastructure.TfData
#            channels: list of channels to plot
#        '''
#        if type(channels) is int:
#            channels = [channels]
#            
#        if not 'list' in data.__class__.__name__.lower():
#            # if a raw data object is passed, first put it into expected list format
#            data = [data]
#            
#        if data[0].__class__.__name__ == 'DataSet':
#            # if DataSet class then undo turning it into a list
#            data = data[0]
#            self.data = data
#            
#            if len(data.time_data_list)>0:
#                self.plot_time_data(data.time_data_list,channels)
#            if len(data.freq_data_list)>0:
#                self.plot_freq_data(data.freq_data_list,channels)
#            if len(data.tf_data_list)>0:
#                self.plot_tf_data(data.tf_data_list,channels)
#            
#        elif data[0].__class__.__name__ == 'TimeData':
#            self.data = datastructure.DataSet()
#            self.data.add_to_dataset(data)
#            self.plot_time_data(data,channels)
#        
#        elif data[0].__class__.__name__  == 'FreqData':
#            self.data = datastructure.DataSet()
#            self.data.add_to_dataset(data)
#            self.plot_freq_data(data,channels)
#            
#        elif data[0].__class__.__name__  == 'TfData':
#            self.data = datastructure.DataSet()
#            self.data.add_to_dataset(data)
#            self.plot_tf_data(data,channels,plot_coherence)
#            
#        self.channels = channels
#            
#            
#            
#    def plot_time_data(self,time_data_list,channels):
#        ### plot time domain data
#        self.timefig, self.timeax = plt.subplots(figsize = (9,5),dpi=100)
#    
#        self.timeax.set_xlabel('Time (s)')
#        self.timeax.set_ylabel('Normalised Amplitude')
#        self.timeax.grid()
#        
#        if channels == 'all':
#            channels = list(range(time_data_list[0].settings.channels))
#            print(channels)
#            
#        count = -1
#        for n_set in range(len(time_data_list)):
#            for n_chan in range(time_data_list[n_set].settings.channels):
#                count += 1
#                if n_chan in channels:
#                    self.timeax.plot(time_data_list[n_set].time_axis,time_data_list[n_set].time_data[:,n_chan],'-',linewidth=1,color = options.set_plot_colours(len(time_data_list)*time_data_list[n_set].settings.channels)[count,:]/255,label='set{},ch{}'.format(n_set,n_chan))
#            
#        self.timeax.legend()
#        
#        plt.show()
#        
#    def plot_freq_data(self,freq_data_list,channels):
#        ### plot frequency domain data
#        self.freqfig, self.freqax = plt.subplots(figsize = (9,5),dpi=100)
#    
#        self.freqax.set_xlabel('Frequency (Hz)')
#        self.freqax.set_ylabel('Amplitude (dB)')
#        self.freqax.grid()
#        
#        if channels == 'all':
#            channels = list(range(freq_data_list[0].settings.channels))
#            
#        count = -1
#        for n_set in range(len(freq_data_list)):
#            for n_chan in range(freq_data_list[n_set].settings.channels):
#                count += 1
#                if n_chan in channels:
#                    self.freqax.plot(freq_data_list[n_set].freq_axis,20*np.log10(np.abs(freq_data_list[n_set].freq_data[:,n_chan])),'-',linewidth=1,color = options.set_plot_colours(len(freq_data_list)*freq_data_list[n_set].settings.channels)[count,:]/255,label='set{},ch{}'.format(n_set,n_chan))
#            
#        self.freqax.legend()
#        
#        plt.show()
#        
#    def plot_tf_data(self,tf_data_list,channels,plot_coherence=True):
#        ### plot transfer function data
#        self.tffig, self.tfax = plt.subplots(figsize = (9,5),dpi=100)
#    
#        self.tfax.set_xlabel('Frequency (Hz)')
#        self.tfax.set_ylabel('Amplitude (dB)')
#        self.tfax.grid()
#        
#        if channels == 'all':
#            channels = list(range(tf_data_list[0].settings.channels-1))
#        
#        count = -1
#        for n_set in range(len(tf_data_list)):
#            for n_chan in range(tf_data_list[n_set].settings.channels-1):
#                count += 1
#                if n_chan in channels:
#                    self.tfax.plot(tf_data_list[n_set].freq_axis,20*np.log10(np.abs(tf_data_list[n_set].tf_data[:,n_chan])),'-',linewidth=1,color = options.set_plot_colours(len(tf_data_list)*tf_data_list[n_set].settings.channels)[count,:]/255,label='set{},ch{}'.format(n_set,n_chan))
#                    if plot_coherence and not np.any(tf_data_list[n_set].tf_coherence == None):
#                        self.tfax.plot(tf_data_list[n_set].freq_axis,20*np.log10(np.abs(tf_data_list[n_set].tf_coherence[:,n_chan])),'--',linewidth=1,color = options.set_plot_colours(len(tf_data_list)*tf_data_list[n_set].settings.channels)[count,:]/255,label='set{},ch{} (coherence)'.format(n_set,n_chan))
#            
#        self.tfax.legend()
#        
#        plt.show()
#        
#        
#

    
    
    


